from abc import ABC, abstractmethod

class TemporalContextLoader(ABC):

    def __init__(self, traffic_dataloader):

        self.weather_data = self.get_weather()
        self.holiday_data = self.get_holiday()
        self.weather_data = self.get_temporal_position()
        pass

    @abstractmethod
    def get_weather(self, arg):
        '''
        Input:
            default weather features shape: (T, D_t)
            detailed weather features shape: (T, N, D_t)
            station info [[lat1, lng1], ....]

        Function to be implemented:
            Func 1 (T, ?, D_t) -> bind/replicate (optional) -> (T, N, D_t)
            Func 2 move Context sample
        '''
        # # func 1 推断天气
        # self.infer_weather()

        # func 2 构造历史context样本
        # # move Context sample
        # self.train_ef_closeness = None
        # self.train_ef_period = None
        # self.train_ef_trend = None
        # self.train_lstm_ef =  None
        # self.test_ef_closeness = None
        # self.test_ef_period = None
        # self.test_ef_trend = None
        # self.test_lstm_ef = None
        # if len(external_feature) > 0:
        #     self.external_move_sample = ST_MoveSample(closeness_len=self.closeness_len,
        #                                     period_len=self.period_len,
        #                                     trend_len=self.trend_len, target_length=0, daily_slots=self.daily_slots)

        #     self.train_ef_closeness, self.train_ef_period, self.train_ef_trend, _ = self.external_move_sample.move_sample(self.train_ef)

        #     self.test_ef_closeness, self.test_ef_period, self.test_ef_trend, _ = self.external_move_sample.move_sample(self.test_ef)


        #     if self.external_lstm_len is not None and self.external_lstm_len > 0:    
        #         self.external_move_sample = ST_MoveSample(closeness_len=self.external_lstm_len,period_len=0,trend_len=0, target_length=0, daily_slots=self.daily_slots)

        #         self.train_lstm_ef, _, _, _ = self.external_move_sample.move_sample(self.train_ef)

        #         self.test_lstm_ef, _, _, _ = self.external_move_sample.move_sample(self.test_ef)

        #     self.train_ef = self.train_ef[-self.train_sequence_len - target_length: -target_length]
        #     self.test_ef = self.test_ef[-self.test_sequence_len - target_length: -target_length]
            
        #     # weather
        #     self.train_lstm_ef = self.train_lstm_ef[-self.train_sequence_len - target_length: -target_length]
        #     self.test_lstm_ef = self.test_lstm_ef[-self.test_sequence_len - target_length: -target_length]
            
        # pass

    # @abstractmethod
    # def infer_weather(sellf, arg):
    #     pass
    
    @abstractmethod
    def get_holiday(sellf, arg):
        '''
        Input:
            parser function or csv file

        Function to be implemented:
            Func 1 (T//daily_slots, 1) -> temporal replicate ->  (T, 1)
        '''
        # parse by api

        # load by file
        pass
    
    # @staticmethod
    @abstractmethod
    def get_temporal_position(self, arg):
        ...



class SpatialContextLoader(ABC):
     
    def __init__(self, traffic_dataloader):
        self.poi = self.get_poi()
        
    @abstractmethod
    def get_poi(sellf, arg):
        '''
        Input:
            parser function or csv file

        Function to be implemented:
            Func 1 (?, N, D_s) -> temporal replicate -> (T, N, D_s)
        '''

        # load by file
        pass


import time
import numpy as np
import mxnet as mx
from UCTB.train.LossFunction import masked_mse_np
from UCTB.evaluation.metric import rmse, mape
from UCTB.preprocess.GraphGenerator import GraphGenerator
from UCTB.preprocess import SplitData
from UCTB.model.STSGCN import construct_model


def configData(args, data_loader, batch_size, config, ctx):
    print("args['normalize']", args.normalize)
    normalizer = data_loader.normalizer
    graph_obj = GraphGenerator(graph='distance', data_loader=data_loader)

    config["num_of_vertices"] = data_loader.station_number
    config["points_per_hour"] = data_loader.closeness_len + data_loader.period_len + data_loader.trend_len
    num_of_vertices = config["num_of_vertices"]
    net = construct_model(config, AM=graph_obj.AM[0])

    model_name = "{}_{}_{}".format(args.dataset, args.city, args.MergeIndex)
    print("model_name:", model_name)

    train_closeness, val_closeness = SplitData.split_data(data_loader.train_closeness, [0.9, 0.1])
    train_period, val_period = SplitData.split_data(data_loader.train_period, [0.9, 0.1])
    train_trend, val_trend = SplitData.split_data(data_loader.train_trend, [0.9, 0.1])
    train_y, val_y = SplitData.split_data(data_loader.train_y, [0.9, 0.1])

    # T, num_node, 1 -> T, 1, num_node
    train_y = train_y.transpose([0, 2, 1])
    val_y = val_y.transpose([0, 2, 1])
    test_y = data_loader.test_y.transpose([0, 2, 1])

    # T, num_node, dimension, 1 -> T, dimension, num_node, 1

    seq_train = np.concatenate([train_trend, train_period, train_closeness], axis=2).transpose([0, 2, 1, 3])
    seq_val = np.concatenate([val_trend, val_period, val_closeness], axis=2).transpose([0, 2, 1, 3])
    seq_test = np.concatenate([data_loader.test_trend, data_loader.test_period, data_loader.test_closeness],
                                  axis=2).transpose([0, 2, 1, 3])


    print(seq_train.shape, seq_val.shape, seq_test.shape)

    loaders = []
    true_values = []
    for item in ["train", "val", "test"]:
        loaders.append(
            mx.io.NDArrayIter(
                eval("seq_{}".format(item)), eval("{}_y".format(item)),
                batch_size=batch_size,
                shuffle=True,
                label_name='label'
            )
        )
        true_values.append(eval("{}_y".format(item)))

    train_loader, val_loader, test_loader = loaders
    _, val_y, test_y = true_values

    global_epoch = 1
    training_samples = len(seq_train)
    global_train_steps = training_samples // batch_size + 1
    all_info = []
    epochs = config['epochs']

    mod = mx.mod.Module(
        net,
        data_names=['data'],
        label_names=['label'],
        context=ctx
    )

    mod.bind(
        data_shapes=[(
            'data',
            (batch_size, config['points_per_hour'], num_of_vertices, 1)
        ), ],
        label_shapes=[(
            'label',
            (batch_size, config['num_for_predict'], num_of_vertices)
        )]
    )

    mod.init_params(initializer=mx.init.Xavier(magnitude=0.0003))
    lr_sch = mx.lr_scheduler.PolyScheduler(
        max_update=global_train_steps * epochs * config['max_update_factor'],
        base_lr=config['learning_rate'],
        pwr=2,
        warmup_steps=global_train_steps
    )

    mod.init_optimizer(
        optimizer=config['optimizer'],
        optimizer_params=(('lr_scheduler', lr_sch),)
    )

    num_of_parameters = 0
    for param_name, param_value in mod.get_params()[0].items():
        # print(param_name, param_value.shape)
        num_of_parameters += np.prod(param_value.shape)
    print("Number of Parameters: {}".format(num_of_parameters), flush=True)

    metric = mx.metric.create(['RMSE', 'MAE'], output_names=['pred_output'])

    if args.plot:
        graph = mx.viz.plot_network(net)
        graph.format = 'png'
        graph.render('graph')
    return model_name, epochs, metric, mod, train_loader, val_loader, test_loader, normalizer, val_y, test_y, all_info


def training(epochs, metric, mod, train_loader, val_loader, test_loader, normalizer, val_y, test_y, all_info):
    global global_epoch
    global_epoch = 1
    lowest_val_loss = np.inf
    for _ in range(epochs):
        t = time.time()
        info = [global_epoch]
        train_loader.reset()
        metric.reset()
        for idx, databatch in enumerate(train_loader):
            # print(databatch,type(databatch))
            mod.forward_backward(databatch)
            mod.update_metric(metric, databatch.label)
            mod.update()
        metric_values = dict(zip(*metric.get()))

        print('training: Epoch: %s, RMSE: %.2f, MAE: %.2f, time: %.2f s' % (
            global_epoch, metric_values['rmse'], metric_values['mae'],
            time.time() - t), flush=True)
        # info.append(metric_values['mae'])
        info.append(metric_values['rmse'])

        val_loader.reset()
        prediction = mod.predict(val_loader)[1].asnumpy()
        # loss = masked_mae_np(val_y, prediction, 0)
        loss = masked_mse_np(val_y, prediction, 0)
        print('validation: Epoch: %s, loss: %.2f, time: %.2f s' % (
            global_epoch, loss, time.time() - t), flush=True)
        info.append(loss)

        if loss < lowest_val_loss:

            test_loader.reset()
            prediction = mod.predict(test_loader)[1].asnumpy()

            prediction = normalizer.inverse_transform(prediction)
            de_norm_test_y = normalizer.inverse_transform(test_y)
            rmse_result = rmse(prediction=prediction.squeeze(), target=de_norm_test_y.squeeze(), threshold=0)
            mape_result = mape(prediction=prediction.squeeze(), target=de_norm_test_y.squeeze(), threshold=0.01)

            print('test: Epoch: {}, MAPE: {:.2f}, RMSE: {:.2f}, '
                  'time: {:.2f}s'.format(
                global_epoch, mape_result, rmse_result, time.time() - t))
            print(flush=True)
            info.extend((mape_result, rmse_result))
            all_info.append(info)
            lowest_val_loss = loss

        global_epoch += 1






import numpy as np
import mxnet as mx
from UCTB.train.LossFunction import huber_loss


def position_embedding(data,
                       input_length, num_of_vertices, embedding_size,
                       temporal=True, spatial=True,
                       init=mx.init.Xavier(magnitude=0.0003), prefix=""):
    '''
    Parameters
    ----------
    data: mx.sym.var, shape is (B, T, N, C)

    input_length: int, length of time series, T

    num_of_vertices: int, N

    embedding_size: int, C

    temporal, spatial: bool, whether equip this type of embeddings

    init: mx.initializer.Initializer

    prefix: str

    Returns
    ----------
    data: output shape is (B, T, N, C)
    '''

    temporal_emb = None
    spatial_emb = None

    if temporal:
        # shape is (1, T, 1, C)
        temporal_emb = mx.sym.var(
            "{}_t_emb".format(prefix),
            shape=(1, input_length, 1, embedding_size),
            init=init
        )
    if spatial:
        # shape is (1, 1, N, C)
        spatial_emb = mx.sym.var(
            "{}_v_emb".format(prefix),
            shape=(1, 1, num_of_vertices, embedding_size),
            init=init
        )

    if temporal_emb is not None:
        data = mx.sym.broadcast_add(data, temporal_emb)
    if spatial_emb is not None:
        data = mx.sym.broadcast_add(data, spatial_emb)

    return data


def gcn_operation(data, adj,
                  num_of_filter, num_of_features, num_of_vertices,
                  activation, prefix=""):
    '''
    graph convolutional operation, a simple GCN we defined in paper

    Parameters
    ----------
    data: mx.sym.var, shape is (3N, B, C)

    adj: mx.sym.var, shape is (3N, 3N)

    num_of_filter: int, C'

    num_of_features: int, C

    num_of_vertices: int, N

    activation: str, {'GLU', 'relu'}

    prefix: str

    Returns
    ----------
    output shape is (3N, B, C')

    '''

    assert activation in {'GLU', 'relu'}

    # shape is (3N, B, C)
    data = mx.sym.dot(adj, data)

    if activation == 'GLU':

        # shape is (3N, B, 2C')
        data = mx.sym.FullyConnected(
            data,
            flatten=False,
            num_hidden=2 * num_of_filter
        )

        # shape is (3N, B, C'), (3N, B, C')
        lhs, rhs = mx.sym.split(data, num_outputs=2, axis=2)

        # shape is (3N, B, C')
        return lhs * mx.sym.sigmoid(rhs)

    elif activation == 'relu':

        # shape is (3N, B, C')
        return mx.sym.Activation(
            mx.sym.FullyConnected(
                data,
                flatten=False,
                num_hidden=num_of_filter
            ), activation
        )


def stsgcm(data, adj,
           filters, num_of_features, num_of_vertices,
           activation, prefix=""):
    '''
    STSGCM, multiple stacked gcn layers with cropping and max operation

    Parameters
    ----------
    data: mx.sym.var, shape is (3N, B, C)

    adj: mx.sym.var, shape is (3N, 3N)

    filters: list[int], list of C'

    num_of_features: int, C

    num_of_vertices: int, N

    activation: str, {'GLU', 'relu'}

    prefix: str

    Returns
    ----------
    output shape is (N, B, C')

    '''
    need_concat = []

    for i in range(len(filters)):
        data = gcn_operation(
            data, adj,
            filters[i], num_of_features, num_of_vertices,
            activation=activation,
            prefix="{}_gcn_{}".format(prefix, i)
        )
        need_concat.append(data)
        num_of_features = filters[i]

    # shape of each element is (1, N, B, C')
    need_concat = [
        mx.sym.expand_dims(
            mx.sym.slice(
                i,
                begin=(num_of_vertices, None, None),
                end=(2 * num_of_vertices, None, None)
            ), 0
        ) for i in need_concat
    ]

    # shape is (N, B, C')
    return mx.sym.max(mx.sym.concat(*need_concat, dim=0), axis=0)


def stsgcl(data, adj,
           T, num_of_vertices, num_of_features, filters,
           module_type, activation, temporal_emb=True, spatial_emb=True,
           prefix=""):
    '''
    STSGCL

    Parameters
    ----------
    data: mx.sym.var, shape is (B, T, N, C)

    adj: mx.sym.var, shape is (3N, 3N)

    T: int, length of time series, T

    num_of_vertices: int, N

    num_of_features: int, C

    filters: list[int], list of C'

    module_type: str, {'sharing', 'individual'}

    activation: str, {'GLU', 'relu'}

    temporal_emb, spatial_emb: bool

    prefix: str

    Returns
    ----------
    output shape is (B, T-2, N, C')
    '''

    assert module_type in {'sharing', 'individual'}

    if module_type == 'individual':
        return sthgcn_layer_individual(
            data, adj,
            T, num_of_vertices, num_of_features, filters,
            activation, temporal_emb, spatial_emb, prefix
        )
    else:
        return sthgcn_layer_sharing(
            data, adj,
            T, num_of_vertices, num_of_features, filters,
            activation, temporal_emb, spatial_emb, prefix
        )


def sthgcn_layer_individual(data, adj,
                            T, num_of_vertices, num_of_features, filters,
                            activation, temporal_emb=True, spatial_emb=True,
                            prefix=""):
    '''
    STSGCL, multiple individual STSGCMs

    Parameters
    ----------
    data: mx.sym.var, shape is (B, T, N, C)

    adj: mx.sym.var, shape is (3N, 3N)

    T: int, length of time series, T

    num_of_vertices: int, N

    num_of_features: int, C

    filters: list[int], list of C'

    activation: str, {'GLU', 'relu'}

    temporal_emb, spatial_emb: bool

    prefix: str

    Returns
    ----------
    output shape is (B, T-2, N, C')
    '''

    # shape is (B, T, N, C)
    data = position_embedding(data, T, num_of_vertices, num_of_features,
                              temporal_emb, spatial_emb,
                              prefix="{}_emb".format(prefix))
    need_concat = []
    for i in range(T - 2):

        # shape is (B, 3, N, C)
        t = mx.sym.slice(data, begin=(None, i, None, None),
                         end=(None, i + 3, None, None))

        # shape is (B, 3N, C)
        t = mx.sym.reshape(t, (-1, 3 * num_of_vertices, num_of_features))

        # shape is (3N, B, C)
        t = mx.sym.transpose(t, (1, 0, 2))

        # shape is (N, B, C')
        t = stsgcm(
            t, adj, filters, num_of_features, num_of_vertices,
            activation=activation,
            prefix="{}_stsgcm_{}".format(prefix, i)
        )

        # shape is (B, N, C')
        t = mx.sym.swapaxes(t, 0, 1)

        # shape is (B, 1, N, C')
        need_concat.append(mx.sym.expand_dims(t, axis=1))

    # shape is (B, T-2, N, C')
    return mx.sym.concat(*need_concat, dim=1)


def sthgcn_layer_sharing(data, adj,
                         T, num_of_vertices, num_of_features, filters,
                         activation, temporal_emb=True, spatial_emb=True,
                         prefix=""):
    '''
    STSGCL, multiple a sharing STSGCM

    Parameters
    ----------
    data: mx.sym.var, shape is (B, T, N, C)

    adj: mx.sym.var, shape is (3N, 3N)

    T: int, length of time series, T

    num_of_vertices: int, N

    num_of_features: int, C

    filters: list[int], list of C'

    activation: str, {'GLU', 'relu'}

    temporal_emb, spatial_emb: bool

    prefix: str

    Returns
    ----------
    output shape is (B, T-2, N, C')
    '''

    # shape is (B, T, N, C)
    data = position_embedding(data, T, num_of_vertices, num_of_features,
                              temporal_emb, spatial_emb,
                              prefix="{}_emb".format(prefix))
    need_concat = []
    for i in range(T - 2):
        # shape is (B, 3, N, C)
        t = mx.sym.slice(data, begin=(None, i, None, None),
                         end=(None, i + 3, None, None))

        # shape is (B, 3N, C)
        t = mx.sym.reshape(t, (-1, 3 * num_of_vertices, num_of_features))

        # shape is (3N, B, C)
        t = mx.sym.swapaxes(t, 0, 1)
        need_concat.append(t)

    # shape is (3N, (T-2)*B, C)
    t = mx.sym.concat(*need_concat, dim=1)

    # shape is (N, (T-2)*B, C')
    t = stsgcm(
        t, adj, filters, num_of_features, num_of_vertices,
        activation=activation,
        prefix="{}_stsgcm".format(prefix)
    )

    # shape is (N, T - 2, B, C)
    t = t.reshape((num_of_vertices, T - 2, -1, filters[-1]))

    # shape is (B, T - 2, N, C)
    return mx.sym.swapaxes(t, 0, 2)


def output_layer(data, num_of_vertices, input_length, num_of_features,
                 num_of_filters=128, predict_length=12):
    '''
    Parameters
    ----------
    data: mx.sym.var, shape is (B, T, N, C)

    num_of_vertices: int, N

    input_length: int, length of time series, T

    num_of_features: int, C

    num_of_filters: int, C'

    predict_length: int, length of predicted time series, T'

    Returns
    ----------
    output shape is (B, T', N)
    '''

    # data shape is (B, N, T, C)
    data = mx.sym.swapaxes(data, 1, 2)

    # (B, N, T * C)
    data = mx.sym.reshape(
        data, (-1, num_of_vertices, input_length * num_of_features)
    )

    # (B, N, C')
    data = mx.sym.Activation(
        mx.sym.FullyConnected(
            data,
            flatten=False,
            num_hidden=num_of_filters
        ), 'relu'
    )

    # (B, N, T')
    data = mx.sym.FullyConnected(
        data,
        flatten=False,
        num_hidden=predict_length
    )

    # (B, T', N)
    data = mx.sym.swapaxes(data, 1, 2)

    return data





def stsgcn(data, adj, label,
           input_length, num_of_vertices, num_of_features,
           filter_list, module_type, activation,
           use_mask=True, mask_init_value=None,
           temporal_emb=True, spatial_emb=True,
           prefix="", rho=1, predict_length=12):
    """

    References:
        - `Spatial-temporal synchronous graph convolutional networks: A new framework for spatial-temporal network data forecasting.
          <https://ojs.aaai.org/index.php/AAAI/article/view/5438>`_.
        - `A Mxnet implementation of the stsgcn model  (Davidham3)
          <https://github.com/Davidham3/STSGCN>`_.

    Args:
        data(mxnet.sym): Input data.
        adj(mxnet.sym): Adjacent matrix.
        label(mxnet.sym): Prediction label.
        input_length(int): Length of input data.
        num_of_vertices(int): Number of vertices in the graph.
        num_of_features(int): Number of features of each vertice.
        filter_list(list): Filters.
        module_type(str): Whether sharing weights.
        activation(str): Choose which activate function.
        use_mask(bool): Whether we use mask.
        mask_init_value(int): Initial value of mask.
        temporal_emb(bool): Whether to use temporal embedding.
        spatial_emb(bool): Whether to use spatial embedding.
        prefix(str): String prefix of mask.
        rho(float): Hyperparameters used to calculate huber loss.
        predict_length(int): Length of prediction.
    """
    '''
    data shape is (B, T, N, C)
    adj shape is (3N, 3N)
    label shape is (B, T, N)
    '''
    if use_mask:
        if mask_init_value is None:
            raise ValueError("mask init value is None!")
        mask = mx.sym.var("{}_mask".format(prefix),
                          shape=(3 * num_of_vertices, 3 * num_of_vertices),
                          init=mask_init_value)
        adj = mask * adj

    for idx, filters in enumerate(filter_list):
        data = stsgcl(data, adj, input_length, num_of_vertices,
                      num_of_features, filters, module_type,
                      activation=activation,
                      temporal_emb=temporal_emb,
                      spatial_emb=spatial_emb,
                      prefix="{}_stsgcl_{}".format(prefix, idx))
        input_length -= 2
        num_of_features = filters[-1]

    # (B, 1, N)
    need_concat = []
    for i in range(predict_length):
        need_concat.append(
            output_layer(
                data, num_of_vertices, input_length, num_of_features,
                num_of_filters=128, predict_length=1
            )
        )
    data = mx.sym.concat(*need_concat, dim=1)

    loss = huber_loss(data, label, rho=rho)
    return mx.sym.Group([loss, mx.sym.BlockGrad(data, name='pred')])



def construct_model(config, AM):

    module_type = config['module_type']
    act_type = config['act_type']
    temporal_emb = config['temporal_emb']
    spatial_emb = config['spatial_emb']
    use_mask = config['use_mask']
    batch_size = config['batch_size']

    num_of_vertices = config['num_of_vertices']
    num_of_features = config['num_of_features']
    points_per_hour = config['points_per_hour']
    num_for_predict = config['num_for_predict']

    adj = AM
    # print("Adj:",adj.shape,adj)
    adj_mx = construct_adj(adj, 3)
    print("The shape of localized adjacency matrix: {}".format(
        adj_mx.shape), flush=True)

    data = mx.sym.var("data")
    label = mx.sym.var("label")
    adj = mx.sym.Variable('adj', shape=adj_mx.shape,
                          init=mx.init.Constant(value=adj_mx.tolist()))
    adj = mx.sym.BlockGrad(adj)
    mask_init_value = mx.init.Constant(value=(adj_mx != 0)
                                       .astype('float32').tolist())

    filters = config['filters']
    first_layer_embedding_size = config['first_layer_embedding_size']
    if first_layer_embedding_size:
        data = mx.sym.Activation(
            mx.sym.FullyConnected(
                data,
                flatten=False,
                num_hidden=first_layer_embedding_size
            ),
            act_type='relu'
        )
    else:
        first_layer_embedding_size = num_of_features
    net = stsgcn(
        data, adj, label,
        points_per_hour, num_of_vertices, first_layer_embedding_size,
        filters, module_type, act_type,
        use_mask, mask_init_value, temporal_emb, spatial_emb,
        prefix="", rho=1, predict_length=1
    )

    assert net.infer_shape(
        data=(batch_size, points_per_hour, num_of_vertices, 1),
        label=(batch_size, num_for_predict, num_of_vertices)
    )[1][1] == (batch_size, num_for_predict, num_of_vertices)
    return net




def get_adjacency_matrix(distance_df_filename, num_of_vertices,
                         type_='connectivity', id_filename=None):
    '''
    Parameters
    ----------
    distance_df_filename: str, path of the csv file contains edges information

    num_of_vertices: int, the number of vertices

    type_: str, {connectivity, distance}

    Returns
    ----------
    A: np.ndarray, adjacency matrix

    '''
    import csv

    A = np.zeros((int(num_of_vertices), int(num_of_vertices)),
                 dtype=np.float32)

    if id_filename:
        with open(id_filename, 'r') as f:
            id_dict = {int(i): idx
                       for idx, i in enumerate(f.read().strip().split('\n'))}
        with open(distance_df_filename, 'r') as f:
            f.readline()
            reader = csv.reader(f)
            for row in reader:
                if len(row) != 3:
                    continue
                i, j, distance = int(row[0]), int(row[1]), float(row[2])
                A[id_dict[i], id_dict[j]] = 1
                A[id_dict[j], id_dict[i]] = 1
        return A

    # Fills cells in the matrix with distances.
    with open(distance_df_filename, 'r') as f:
        f.readline()
        reader = csv.reader(f)
        for row in reader:
            if len(row) != 3:
                continue
            i, j, distance = int(row[0]), int(row[1]), float(row[2])
            if type_ == 'connectivity':
                A[i, j] = 1
                A[j, i] = 1
            elif type == 'distance':
                A[i, j] = 1 / distance
                A[j, i] = 1 / distance
            else:
                raise ValueError("type_ error, must be "
                                 "connectivity or distance!")
    return A


def construct_adj(A, steps):
    '''
    construct a bigger adjacency matrix using the given matrix

    Parameters
    ----------
    A: np.ndarray, adjacency matrix, shape is (N, N)

    steps: how many times of the does the new adj mx bigger than A

    Returns
    ----------
    new adjacency matrix: csr_matrix, shape is (N * steps, N * steps)
    '''
    N = len(A)
    adj = np.zeros([N * steps] * 2)

    for i in range(steps):
        adj[i * N: (i + 1) * N, i * N: (i + 1) * N] = A

    for i in range(N):
        for k in range(steps - 1):
            adj[k * N + i, (k + 1) * N + i] = 1
            adj[(k + 1) * N + i, k * N + i] = 1

    for i in range(len(adj)):
        adj[i, i] = 1

    return adj






import os
import wget
import pickle
import tarfile
import numpy as np


class DataSet(object):
    """An object storing basic data from a formatted pickle file.
    See also `Build your own datasets <./md_file/tutorial.html>`_.
    Args:
        dataset (str): A string containing path of the dataset pickle file or a string of name of the dataset.
        city (str or ``None``): ``None`` if dataset is file path, or a string of name of the city. Default: ``None``
        data_dir (str or ``None``): The dataset directory. If set to ``None``, a directory will be created.
            If ``dataset`` is file path, ``data_dir`` should be ``None`` too. Default: ``None``
    Attributes:
        data (dict): The data directly from the pickle file. ``data`` may have a ``data['contribute_data']`` dict to
            store supplementary data.
        time_range (list): From ``data['TimeRange']`` in the format of [YYYY-MM-DD, YYYY-MM-DD] indicating the time
            range of the data.
        time_fitness (int): From ``data['TimeFitness']`` indicating how many minutes is a single time slot.
        node_traffic (np.ndarray): Data recording the main stream data of the nodes in during the time range.
            From ``data['Node']['TrafficNode']`` with shape as [time_slot_num, node_num].
        node_monthly_interaction (np.ndarray): Data recording the monthly interaction of pairs of nodes.
            Its shape is [month_num, node_num, node_num].It's from ``data['Node']['TrafficMonthlyInteraction']``
            and is used to build interaction graph.
            Its an optional attribute and can be set as an empty list if interaction graph is not needed.
        node_station_info (dict): A dict storing the coordinates of nodes. It shall be formatted as {id (may be
            arbitrary): [id (when sorted, should be consistant with index of ``node_traffic``), latitude, longitude,
            other notes]}. It's from ``data['Node']['StationInfo']`` and is used to build distance graph.
            Its an optional attribute and can be set as an empty list if distance graph is not needed.
        MergeIndex(int): A int number that used to adjust the granularity of the dataset, the granularity of the new
            dataset is time_fitness*MergeIndex. default: 1
        MergeWay(str):can be `sum` and `average`.  default: ``sum
    """

    def __init__(self, dataset, MergeIndex, MergeWay, city=None, data_dir=None):
        self.dataset = dataset
        self.city = city
        self.MergeIndex = int(MergeIndex)
        self.MergeWay = MergeWay.lower()

        if data_dir is None:
            data_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'data')

        if os.path.isdir(data_dir) is False:
            os.makedirs(data_dir)

        if self.city is not None:
            pkl_file_name = os.path.join(data_dir, '{}_{}.pkl'.format(self.dataset, self.city))
        else:
            pkl_file_name = self.dataset

        if os.path.isfile(pkl_file_name) is False:
            try:
                tar_file_name = os.path.join(data_dir, '{}_{}.tar.gz'.format(self.dataset, self.city))
                if os.path.isfile(tar_file_name) is False:
                    print('Downloading data into', data_dir)
                    wget.download('https://github.com/Di-Chai/UCTB_Data/blob/master/%s_%s.tar.gz?raw=true' %
                                  (dataset, city), tar_file_name)
                    print('Download succeed')
                else:
                    print('Found', tar_file_name)
                tar = tarfile.open(tar_file_name, "r:gz")
                file_names = tar.getnames()
                for file_name in file_names:
                    tar.extract(file_name, data_dir)
                tar.close()
                os.remove(tar_file_name)
            except Exception as e:
                print(e)
                raise FileExistsError('Download Failed')

        with open(pkl_file_name, 'rb') as f:
            self.data = pickle.load(f)

        # merge data
        if self.MergeIndex > 1:
            self.data['TimeFitness'] = int(self.data['TimeFitness'] * self.MergeIndex)
            if len(self.data['Node']['TrafficNode']) > 0:
                self.data['Node']['TrafficNode'] = self.merge_data(self.data['Node']['TrafficNode'], "node")
            if len(self.data['Grid']['TrafficGrid']) > 0:
                self.data['Grid']['TrafficGrid'] = self.merge_data(self.data['Grid']['TrafficGrid'], "grid")
            if len(self.data['ExternalFeature']['Weather']) > 0:
                self.data['ExternalFeature']['Weather'] = self.merge_data(self.data['ExternalFeature']['Weather'],
                                                                          "node")

        self.time_range = self.data['TimeRange']
        self.time_fitness = self.data['TimeFitness']

        self.node_traffic = self.data['Node']['TrafficNode']
        self.node_monthly_interaction = self.data['Node']['TrafficMonthlyInteraction']
        self.node_station_info = self.data['Node']['StationInfo']

        self.grid_traffic = self.data['Grid']['TrafficGrid']
        self.grid_lat_lng = self.data['Grid']['GridLatLng']

        self.external_feature_weather = self.data['ExternalFeature']['Weather']

    def merge_data(self, data, dataType):
        if self.MergeWay == "sum":
            func = np.sum
        elif self.MergeWay == "average":
            func = np.mean
        elif self.MergeWay == "max":
            func = np.max
        else:
            raise ValueError("Parameter MerWay should be sum or average")
        if data.shape[0] % self.MergeIndex is not 0:
            raise ValueError("time_slots % MergeIndex should be zero")

        if dataType.lower() == "node":
            new = np.zeros((data.shape[0] // self.MergeIndex, data.shape[1]), dtype=np.float32)
            for new_ind, ind in enumerate(range(0, data.shape[0], self.MergeIndex)):
                new[new_ind, :] = func(data[ind:ind + self.MergeIndex, :], axis=0)
        elif dataType.lower() == "grid":
            new = np.zeros((data.shape[0] // self.MergeIndex, data.shape[1], data.shape[2]), dtype=np.float32)
            for new_ind, ind in enumerate(range(0, data.shape[0], self.MergeIndex)):
                new[new_ind, :, :] = func(data[ind:ind + self.MergeIndex, :, :], axis=0)
        return new


from dateutil.parser import parse
import pickle
from datetime import timedelta
import os

def print_dic_info(dic, dic_name, tag=''):
    for k in dic:
        print(tag, end='')
        if type(dic[k])==type({}):
            print(f'{dic_name}[{k}]:{type(dic[k])}'+'{')
            print_dic_info(dic[k], f'{dic_name}[{k}]', tag=tag+'\t')
            print('}', end='')
        elif 'numpy' in str(type(dic[k])):
            print(f'{dic_name}[{k}]:{type(dic[k])}  (shape={dic[k].shape})', end='')
        elif type(dic[k])== type([]):
            lis = dic[k]
            s = f'({len(lis)}, {len(lis[0])})' if len(lis)>0 and type(lis[0])==type([]) else f'{len(lis)}'
            print(f'{dic_name}[{k}]:{type(dic[k])}  (len={s})', end='')
        else:
            print(f'{dic_name}[{k}]:{type(dic[k])}', end='')
        print()

def get_timedelta(dic):
    return timedelta(days=0, seconds=0, microseconds=0,milliseconds=0, minutes=dic['TimeFitness'], hours=0, weeks=0)


def build_uctb_dataset(traffic_node, time_fitness, node_station_info, time_range, dataset_name, city,
    traffic_monthly_interaction=None, external_feature_weather=None, node_poi=None, 
    traffic_grid=None, grid_lat_lng=None, gird_poi=None, print_dataset=False, output_dir=None):
    """
    build and return the uctb dataset in dic format
    necessary:
        time_fitness
        time_range
        Node
            traffic_node
            node_satation_info          
        
    optional: 
        Node
            traffic_monthly_interaction
            poi
        Grid
            traffic_grid
            gird_lat_lng
            poi
        ExternalFeature
            Weather
    """
    dataset = {'TimeRange':time_range, 'TimeFitness':time_fitness, 'Node':{'TrafficNode':traffic_node, 'StationInfo':node_station_info},
                    'Grid':{}, 'ExternalFeature':{}, 'LenTimeSlots':traffic_node.shape[0]}

    # make sure no data missing in traffic node
    beg_dt = parse(dataset['TimeRange'][0])
    assert beg_dt+(dataset['Node']['TrafficNode'].shape[0])*get_timedelta(dataset) == parse(dataset['TimeRange'][1])

    dataset['Node']['TrafficMonthlyInteraction'] = traffic_monthly_interaction
    dataset['Grid']['TrafficGrid'] = traffic_grid
    dataset['Grid']['GridLatLng'] = grid_lat_lng
    dataset['ExternalFeature']['Weather'] = [] if external_feature_weather is None else external_feature_weather
    
    if node_poi is not None:
        dataset['Node']['POI'] = node_poi
        
    if gird_poi is not None:
        dataset['Grid']['POI'] = gird_poi
    
    if print_dataset:
        print_dic_info(dataset, 'dataset')

    if output_dir is None:
        pkl_file_name = '{}_{}.pkl'.format(dataset_name, city)
    else:
        pkl_file_name = os.path.join(output_dir, '{}_{}.pkl'.format(dataset_name, city))

    with open(pkl_file_name, 'wb') as f:
        pickle.dump(dataset, f)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf

from tensorflow.contrib.rnn import RNNCell


class DCGRUCell(RNNCell):
    """Graph Convolution Gated Recurrent Unit cell.
    """

    def call(self, inputs, **kwargs):
        pass

    def compute_output_shape(self, input_shape):
        pass

    def __init__(self, num_units, input_dim, num_graphs, supports, max_diffusion_step, num_node, num_proj=None,
                 activation=tf.nn.tanh, reuse=None, use_gc_for_ru=True, name=None):
        """

        :param num_units:
        :param adj_mx:
        :param max_diffusion_step:
        :param num_node:
        :param input_size:
        :param num_proj:
        :param activation:
        :param reuse:
        :param filter_type: "laplacian", "random_walk", "dual_random_walk".
        :param use_gc_for_ru: whether to use Graph convolution to calculate the reset and update gates.
        """
        super(DCGRUCell, self).__init__(_reuse=reuse)
        self._activation = activation
        self._num_node = num_node
        self._input_dim = input_dim
        self._num_graphs = num_graphs
        self._num_proj = num_proj
        self._num_units = num_units
        self._max_diffusion_step = max_diffusion_step
        self._use_gc_for_ru = use_gc_for_ru
        self._supports = supports
        self._num_diff_matrix = supports.get_shape()[0].value
        self._name = name

    @property
    def state_size(self):
        return self._num_node * self._num_units

    @property
    def output_size(self):
        output_size = self._num_node * self._num_units
        if self._num_proj is not None:
            output_size = self._num_node * self._num_proj
        return output_size

    def __call__(self, inputs, state, scope=None):
        """Gated recurrent unit (GRU) with Graph Convolution.
        :param inputs: (B, num_node * input_dim)

        :return
        - Output: A `2-D` tensor with shape `[batch_size x self.output_size]`.
        - New state: Either a single `2-D` tensor, or a tuple of tensors matching
            the arity and shapes of `state`
        """
        with tf.variable_scope(scope or "dcgru_cell"):
            with tf.variable_scope("gates"):  # Reset gate and update gate.
                output_size = 2 * self._num_units
                # We start with bias of 1.0 to not reset and not update.
                if self._use_gc_for_ru:
                    fn = self._gconv
                else:
                    fn = self._fc
                value = tf.nn.sigmoid(fn(inputs, state, output_size, bias_start=1.0))
                value = tf.reshape(value, (-1, self._num_node, output_size))
                r, u = tf.split(value=value, num_or_size_splits=2, axis=-1)
                r = tf.reshape(r, (-1, self._num_node * self._num_units))
                u = tf.reshape(u, (-1, self._num_node * self._num_units))
            with tf.variable_scope("candidate"):
                c = self._gconv(inputs, r * state, self._num_units)
                if self._activation is not None:
                    c = self._activation(c)
            output = new_state = u * state + (1 - u) * c
            if self._num_proj is not None:
                with tf.variable_scope("projection"):
                    w = tf.get_variable('w', shape=(self._num_units, self._num_proj))
                    output = tf.reshape(new_state, shape=(-1, self._num_units))
                    output = tf.reshape(tf.matmul(output, w), shape=(-1, self.output_size))
        return output, new_state

    @staticmethod
    def _concat(x, x_):
        x_ = tf.expand_dims(x_, 0)
        return tf.concat([x, x_], axis=0)

    def _fc(self, inputs, state, output_size, bias_start=0.0):
        dtype = inputs.dtype
        batch_size = inputs.get_shape()[0].value
        inputs = tf.reshape(inputs, (batch_size * self._num_node, -1))
        state = tf.reshape(state, (batch_size * self._num_node, -1))
        inputs_and_state = tf.concat([inputs, state], axis=-1)
        input_size = inputs_and_state.get_shape()[-1].value
        weights = tf.get_variable(
            'weights', [input_size, output_size], dtype=dtype,
            initializer=tf.contrib.layers.xavier_initializer())
        value = tf.nn.sigmoid(tf.matmul(inputs_and_state, weights))
        biases = tf.get_variable("biases", [output_size], dtype=dtype,
                                 initializer=tf.constant_initializer(bias_start, dtype=dtype))
        value = tf.nn.bias_add(value, biases)
        return value

    def _gconv(self, inputs, state, output_size, bias_start=0.0):
        """Graph convolution between input and the graph matrix.

        :param args: a 2D Tensor or a list of 2D, batch x n, Tensors.
        :param output_size:
        :param bias:
        :param bias_start:
        :param scope:
        :return:
        """
        # Reshape input and state to (batch_size, num_node, input_dim/state_dim)
        last_dim = inputs.get_shape()[-1].value
        inputs = tf.reshape(inputs, (-1, self._num_node, int(last_dim / self._num_node)))
        state = tf.reshape(state, (-1, self._num_node, self._num_units))
        inputs_and_state = tf.concat([inputs, state], axis=2)
        input_size = inputs_and_state.get_shape()[2].value
        dtype = inputs.dtype

        x = inputs_and_state
        x0 = tf.transpose(x, perm=[1, 2, 0])  # (num_node, total_arg_size, batch_size)
        x0 = tf.reshape(x0, shape=[self._num_node, -1])
        x = tf.expand_dims(x0, axis=0)

        scope = tf.get_variable_scope()
        with tf.variable_scope(scope.name + (self.name or ''), reuse=False):
            if self._max_diffusion_step == 0:
                pass
            else:
                for index in range(self._num_diff_matrix):
                    x1 = tf.matmul(self._supports[index], x0)
                    x = self._concat(x, x1)

                    for k in range(2, self._max_diffusion_step + 1):
                        x2 = 2 * tf.matmul(self._supports[index], x1) - x0
                        x = self._concat(x, x2)
                        x1, x0 = x2, x1

            num_matrices = self._num_diff_matrix * self._max_diffusion_step + 1  # Adds for x itself.
            x = tf.reshape(x, shape=[num_matrices, self._num_node, input_size, -1])
            x = tf.transpose(x, perm=[3, 1, 2, 0])  # (batch_size, num_node, input_size, order)
            x = tf.reshape(x, shape=[-1, input_size * num_matrices])

            weights = tf.get_variable(
                'weights', [input_size * num_matrices, output_size], dtype=dtype,
                initializer=tf.contrib.layers.xavier_initializer())
            x = tf.matmul(x, weights)  # (batch_size * self._num_node, output_size)

            biases = tf.get_variable("biases", [output_size], dtype=dtype,
                                     initializer=tf.constant_initializer(bias_start, dtype=dtype))
            x = tf.nn.bias_add(x, biases)
        # Reshape res back to 2D: (batch_size, num_node, state_dim) -> (batch_size, num_node * state_dim)
        return tf.reshape(x, [-1, self._num_node * output_size])



import numpy as np

def rmse(prediction, target, threshold=None):
    """
    Args:
        prediction(ndarray): prediction with shape [batch_size, ...]
        target(ndarray): same shape with prediction, [batch_size, ...]
        threshold(float): data smaller or equal to threshold in target
            will be removed in computing the rmse
    """
    if threshold is None:
        return np.sqrt(np.mean(np.square(prediction - target)))
    else:
        return np.sqrt(np.dot(np.square(prediction - target).reshape([1, -1]),
                              target.reshape([-1, 1]) > threshold) / np.sum(target > threshold))[0][0]


def mape(prediction, target, threshold=0):
    """
    Args:
        prediction(ndarray): prediction with shape [batch_size, ...]
        target(ndarray): same shape with prediction, [batch_size, ...]
        threshold(float): data smaller than threshold in target will be removed in computing the mape,
            the smallest threshold in mape is zero.
    """
    assert threshold > 0
    return (np.dot((np.abs(prediction - target) / (target + (1 - (target > threshold)))).reshape([1, -1]),
                   target.reshape([-1, 1]) > threshold) / np.sum(target > threshold))[0, 0]







import os
from time import time
import numpy as np
import shutil

import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data

from UCTB.preprocess import SplitData
# from tensorboardX import SummaryWriter
from UCTB.preprocess.preprocessor import normalization
from UCTB.train.LossFunction import masked_mape, masked_mae, masked_rmse, masked_mse


def load_data(data_loader, DEVICE, batch_size, shuffle=True):
    '''
    这个是为PEMS的数据准备的函数
    将x,y都处理成归一化到[-1,1]之前的数据;
    每个样本同时包含所有监测点的数据，所以本函数构造的数据输入时空序列预测模型；
    该函数会把hour, day, week的时间串起来；
    注： 从文件读入的数据，x是最大最小归一化的，但是y是真实值
    这个函数转为mstgcn，astgcn设计，返回的数据x都是通过减均值除方差进行归一化的，y都是真实值
    :param graph_signal_matrix_filename: str
    :param num_of_hours: int
    :param num_of_days: int
    :param num_of_weeks: int
    :param DEVICE:
    :param batch_size: int
    :return:
    three DataLoaders, each dataloader contains:
    test_x_tensor: (B, N_nodes, in_feature, T_input)
    test_decoder_input_tensor: (B, N_nodes, T_output)
    test_target_tensor: (B, N_nodes, T_output)

    '''
    # split data
    train_closeness, val_closeness = SplitData.split_data(
        data_loader.train_closeness, [0.9, 0.1])
    train_period, val_period = SplitData.split_data(
        data_loader.train_period, [0.9, 0.1])
    train_trend, val_trend = SplitData.split_data(
        data_loader.train_trend, [0.9, 0.1])
    train_y, val_y = SplitData.split_data(data_loader.train_y, [0.9, 0.1])

    train_x = np.concatenate(
        [train_trend, train_period, train_closeness], axis=2).transpose([0, 1, 3, 2])
    train_target = train_y

    val_x = np.concatenate(
        [val_trend, val_period, val_closeness], axis=2).transpose([0, 1, 3, 2])
    val_target = val_y

    test_x = np.concatenate([data_loader.test_trend, data_loader.test_period,
                            data_loader.test_closeness], axis=2).transpose([0, 1, 3, 2])
    test_target = data_loader.test_y



    print("train_x", train_x.shape)
    print("val_x", val_x.shape)
    print("test_x", test_x.shape)

    # ------- train_loader -------
    train_x_tensor = torch.from_numpy(train_x).type(
        torch.FloatTensor).to(DEVICE)  # (B, N, F, T)
    train_target_tensor = torch.from_numpy(train_target).type(
        torch.FloatTensor).to(DEVICE)  # (B, N, T)

    train_dataset = torch.utils.data.TensorDataset(
        train_x_tensor, train_target_tensor)

    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=batch_size, shuffle=shuffle)

    # ------- val_loader -------
    val_x_tensor = torch.from_numpy(val_x).type(
        torch.FloatTensor).to(DEVICE)  # (B, N, F, T)
    val_target_tensor = torch.from_numpy(val_target).type(
        torch.FloatTensor).to(DEVICE)  # (B, N, T)

    val_dataset = torch.utils.data.TensorDataset(
        val_x_tensor, val_target_tensor)

    val_loader = torch.utils.data.DataLoader(
        val_dataset, batch_size=batch_size, shuffle=False)

    # ------- test_loader -------
    test_x_tensor = torch.from_numpy(test_x).type(
        torch.FloatTensor).to(DEVICE)  # (B, N, F, T)
    test_target_tensor = torch.from_numpy(test_target).type(
        torch.FloatTensor).to(DEVICE)  # (B, N, T)

    test_dataset = torch.utils.data.TensorDataset(
        test_x_tensor, test_target_tensor)

    test_loader = torch.utils.data.DataLoader(
        test_dataset, batch_size=batch_size, shuffle=False)

    # print
    print('train:', train_x_tensor.size(), train_target_tensor.size())
    print('val:', val_x_tensor.size(), val_target_tensor.size())
    print('test:', test_x_tensor.size(), test_target_tensor.size())

    return train_loader, train_target_tensor, val_loader, val_target_tensor, test_loader, test_target_tensor


def train_main(training_config, params_path, DEVICE, net, val_loader, train_loader, graph_signal_matrix_filename):
    learning_rate = float(training_config['learning_rate'])
    epochs = int(training_config['epochs'])
    start_epoch = int(training_config['start_epoch'])
    batch_size = int(training_config['batch_size'])
    num_of_weeks = int(training_config['num_of_weeks'])
    num_of_days = int(training_config['num_of_days'])
    num_of_hours = int(training_config['num_of_hours'])
    time_strides = num_of_hours
    nb_chev_filter = int(training_config['nb_chev_filter'])
    nb_time_filter = int(training_config['nb_time_filter'])
    in_channels = int(training_config['in_channels'])
    nb_block = int(training_config['nb_block'])
    K = int(training_config['K'])
    loss_function = training_config['loss_function']
    metric_method = training_config['metric_method']
    missing_value = float(training_config['missing_value'])
    if (start_epoch == 0) and (not os.path.exists(params_path)):
        os.makedirs(params_path)
        print('create params directory %s' % (params_path))
    elif (start_epoch == 0) and (os.path.exists(params_path)):
        shutil.rmtree(params_path)
        os.makedirs(params_path)
        print('delete the old one and create params directory %s' % (params_path))
    elif (start_epoch > 0) and (os.path.exists(params_path)):
        print('train from params directory %s' % (params_path))
    else:
        raise SystemExit('Wrong type of model!')

    print('param list:')
    print('CUDA\t', DEVICE)
    print('in_channels\t', in_channels)
    print('nb_block\t', nb_block)
    print('nb_chev_filter\t', nb_chev_filter)
    print('nb_time_filter\t', nb_time_filter)
    print('time_strides\t', time_strides)
    print('batch_size\t', batch_size)
    print('graph_signal_matrix_filename\t', graph_signal_matrix_filename)
    print('start_epoch\t', start_epoch)
    print('epochs\t', epochs)
    masked_flag = 0
    criterion = nn.L1Loss().to(DEVICE)
    criterion_masked = masked_mae
    if loss_function == 'masked_mse':
        criterion_masked = masked_mse  # nn.MSELoss().to(DEVICE)
        masked_flag = 1
    elif loss_function == 'masked_mae':
        criterion_masked = masked_mae
        masked_flag = 1
    elif loss_function == 'mae':
        criterion = nn.L1Loss().to(DEVICE)
        masked_flag = 0
    elif loss_function == 'rmse':
        criterion = nn.MSELoss().to(DEVICE)
        masked_flag = 0
    optimizer = optim.Adam(net.parameters(), lr=learning_rate)
    # sw = SummaryWriter(logdir=params_path, flush_secs=5)
    print(net)

    print('Net\'s state_dict:')
    total_param = 0
    for param_tensor in net.state_dict():
        print(param_tensor, '\t', net.state_dict()[param_tensor].size())
        total_param += np.prod(net.state_dict()[param_tensor].size())
    print('Net\'s total params:', total_param)

    print('Optimizer\'s state_dict:')
    for var_name in optimizer.state_dict():
        print(var_name, '\t', optimizer.state_dict()[var_name])

    global_step = 0
    best_epoch = 0
    best_val_loss = np.inf

    start_time = time()

    if start_epoch > 0:

        params_filename = os.path.join(
            params_path, 'epoch_%s.params' % start_epoch)

        net.load_state_dict(torch.load(params_filename))

        print('start epoch:', start_epoch)

        print('load weight from: ', params_filename)

    # train model
    for epoch in range(start_epoch, epochs):

        params_filename = os.path.join(params_path, 'epoch_%s.params' % epoch)

        if masked_flag:
            val_loss = compute_val_loss_mstgcn(
                net, val_loader, criterion_masked, masked_flag, missing_value, epoch)
        else:
            val_loss = compute_val_loss_mstgcn(
                net, val_loader, criterion, masked_flag, missing_value, epoch)

        if val_loss < best_val_loss:
            best_val_loss = val_loss
            best_epoch = epoch
            torch.save(net.state_dict(), params_filename)
            print('save parameters to file: %s' % params_filename)

        net.train()  # ensure dropout layers are in train mode

        for batch_index, batch_data in enumerate(train_loader):

            encoder_inputs, labels = batch_data

            optimizer.zero_grad()

            outputs = net(encoder_inputs)

            if masked_flag:
                loss = criterion_masked(outputs, labels, missing_value)
            else:
                loss = criterion(outputs, labels)

            loss.backward()

            optimizer.step()

            training_loss = loss.item()

            global_step += 1

            # sw.add_scalar('training_loss', training_loss, global_step)

            if global_step % 10 == 0:

                print('global step: %s, training loss: %.2f, time: %.2fs' %
                      (global_step, training_loss, time() - start_time))

    print('best epoch:', best_epoch)
    return best_epoch


def predict_main(net, global_step, data_loader, data_target_tensor, _mean, _std, params_path):
    '''

    :param global_step: int
    :param data_loader: torch.utils.data.utils.DataLoader
    :param data_target_tensor: tensor
    :param mean: (1, 1, 3, 1)
    :param std: (1, 1, 3, 1)
    :param type: string
    :return:
    '''

    params_filename = os.path.join(
        params_path, 'epoch_%s.params' % global_step)
    print('load weight from:', params_filename)

    net.load_state_dict(torch.load(params_filename))

    net.train(False)  # ensure dropout layers are in test mode

    with torch.no_grad():

        data_target_tensor = data_target_tensor.cpu().numpy()

        loader_length = len(data_loader)  # nb of batch

        prediction = []  # 存储所有batch的output

        input = []  # 存储所有batch的input

        for batch_index, batch_data in enumerate(data_loader):

            encoder_inputs, labels = batch_data

            # (batch, T', 1)
            input.append(encoder_inputs[:, :, 0:1].cpu().numpy())

            outputs = net(encoder_inputs)

            prediction.append(outputs.detach().cpu().numpy())

            if batch_index % 100 == 0:
                print('predicting data set batch %s / %s' %
                      (batch_index + 1, loader_length))

        input = np.concatenate(input, 0)

        prediction = np.concatenate(prediction, 0)  # (batch, T', 1)

        return prediction
        


def compute_val_loss_mstgcn(net, val_loader, criterion,  masked_flag,missing_value, epoch, limit=None):
    '''
    for rnn, compute mean loss on validation set
    :param net: model
    :param val_loader: torch.utils.data.utils.DataLoader
    :param criterion: torch.nn.MSELoss
    :param sw: tensorboardX.SummaryWriter
    :param global_step: int, current global_step
    :param limit: int,
    :return: val_loss
    '''

    net.train(False)  # ensure dropout layers are in evaluation mode

    with torch.no_grad():

        val_loader_length = len(val_loader)  # nb of batch

        tmp = []  # 记录了所有batch的loss

        for batch_index, batch_data in enumerate(val_loader):
            encoder_inputs, labels = batch_data
            outputs = net(encoder_inputs)
            if masked_flag:
                loss = criterion(outputs, labels, missing_value)
            else:
                loss = criterion(outputs, labels)

            tmp.append(loss.item())
            if batch_index % 100 == 0:
                print('validation batch %s / %s, loss: %.2f' % (batch_index + 1, val_loader_length, loss.item()))
            if (limit is not None) and batch_index >= limit:
                break

        validation_loss = sum(tmp) / len(tmp)
        # sw.add_scalar('validation_loss', validation_loss, epoch)
    return validation_loss



def re_normalization(x, mean, std):
    x = x * std + mean
    return x


from scipy import stats


class EarlyStopping(object):
    """Early stop if a span of newest records are not better than the current best record.

    Args:
        patience (int): The span of checked newest records.

    Attributes:
        __record_list (list): List of records.
        __best (float): The current best record.
        __patience (int): The span of checked newest records.
        __p (int): The number of newest records that are worse than the current best record.

    """
    def __init__(self, patience):
        self.__record_list = []
        self.__best = None
        self.__patience = patience
        self.__p = 0

    def stop(self, new_value):
        """Append the new record to the record list
        and check if the number of new records than are worse than the best records exceeds the limit.

        Args:
            new_value (float): The new record generated by the newest model.

        Returns:
            bool: ``True`` if the number of new records than are worse than the best records exceeds the limit and
            triggers early stop, otherwise ``False``.
        """
        self.__record_list.append(new_value)
        if self.__best is None or new_value < self.__best:
            self.__best = new_value
            self.__p = 0
            return False
        else:
            if self.__p < self.__patience:
                self.__p += 1
                return False
            else:
                return True


class EarlyStoppingTTest(object):
    """Early Stop by t-test.

    T-test is a two-sided test for the null hypothesis that 2 independent samples
    have identical average (expected) values. This method takes two intervals according to ``length``
    in the record list and see if they have identical average values. If so, do early stop.

    Args:
        length (int): The length of checked interval.
        p_value_threshold (float): The p-value threshold to decide whether to do early stop.

    Attributes:
        __record_list (list): List of records.
        __best (float): The current best record.
        __test_length (int): The length of checked interval.
        __p_value_threshold (float): The p-value threshold to decide whether to do early stop.
    """
    def __init__(self, length, p_value_threshold):
        self.__record_list = []
        self.__best = None
        self.__test_length = length
        self.__p_value_threshold = p_value_threshold

    def stop(self, new_value):
        """
        Take two intervals in the record list to do t-test.

        Args:
            new_value (float): The new record generated by the newest model.

        Returns:
            bool: ``True`` if p value of t-test is smaller than threshold and
            triggers early stop, otherwise ``False``.
        """
        self.__record_list.append(new_value)
        if len(self.__record_list) >= (self.__test_length * 2):
            lossTTest = stats.ttest_ind(self.__record_list[-self.__test_length:],
                                        self.__record_list[-self.__test_length * 2:-self.__test_length], equal_var=False)
            ttest = lossTTest[0]
            pValue = lossTTest[1]
            print('ttest:', ttest, 'pValue', pValue)
            if pValue > self.__p_value_threshold or ttest > 0:
                return True
            else:
                return False
        else:
            return False



import copy
import pickle
import os
import numpy as np


def save_predict_in_dataset(data_loader, predict_val, method):
    data_dir = os.path.join(os.path.dirname(
        os.path.dirname(os.path.abspath(__file__))), 'data')

    original_file = os.path.join(data_dir, "{}_{}.pkl".format(
        data_loader.dataset.dataset, data_loader.dataset.city))
    file_name = os.path.join(data_dir, "{}_{}_pred.pkl".format(
        data_loader.dataset.dataset, data_loader.dataset.city))

    if os.path.exists(file_name):
        with open(file_name, "rb") as fp:
            pred_data = pickle.load(fp)
    else:
        with open(original_file, "rb") as fp:
            pred_data = pickle.load(fp)
            pred_data['Pred'] = {}

    loader_id = data_loader.loader_id
    if loader_id not in pred_data["Pred"].keys():
        pred_data['Pred'][loader_id] = {}

    pred_data['Pred'][loader_id]["GroundTruth"] = np.squeeze(
        data_loader.test_y)

    if loader_id.endswith("N"):
        # use NodeTrafficLoader
        pred_data['Pred'][loader_id][method] = {}
        pred_data['Pred'][loader_id][method]["traffic_data_index"] = data_loader.traffic_data_index
        pred_data['Pred'][loader_id][method]["TrafficNode"] = np.squeeze(
            predict_val)

    if loader_id.endswith("G"):
        # use GridTrafficLoader
        pred_data['Pred'][loader_id][method] = {}
        pred_data['Pred'][loader_id][method]["TrafficGrid"] = np.squeeze(predict_val)

    with open(file_name, "wb") as fp:
        pickle.dump(pred_data, fp)



# -*- coding:utf-8 -*-
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import torch.utils.data


class Spatial_Attention_layer(nn.Module):
    '''
    compute spatial attention scores
    '''
    def __init__(self, DEVICE, in_channels, num_of_vertices, num_of_timesteps):
        super(Spatial_Attention_layer, self).__init__()
        self.W1 = nn.Parameter(torch.FloatTensor(num_of_timesteps).to(DEVICE))
        self.W2 = nn.Parameter(torch.FloatTensor(in_channels, num_of_timesteps).to(DEVICE))
        self.W3 = nn.Parameter(torch.FloatTensor(in_channels).to(DEVICE))
        self.bs = nn.Parameter(torch.FloatTensor(1, num_of_vertices, num_of_vertices).to(DEVICE))
        self.Vs = nn.Parameter(torch.FloatTensor(num_of_vertices, num_of_vertices).to(DEVICE))


    def forward(self, x):
        '''
        :param x: (batch_size, N, F_in, T)
        :return: (B,N,N)
        '''

        lhs = torch.matmul(torch.matmul(x, self.W1), self.W2)  # (b,N,F,T)(T)->(b,N,F)(F,T)->(b,N,T)

        rhs = torch.matmul(self.W3, x).transpose(-1, -2)  # (F)(b,N,F,T)->(b,N,T)->(b,T,N)

        product = torch.matmul(lhs, rhs)  # (b,N,T)(b,T,N) -> (B, N, N)

        S = torch.matmul(self.Vs, torch.sigmoid(product + self.bs))  # (N,N)(B, N, N)->(B,N,N)

        S_normalized = F.softmax(S, dim=1)

        return S_normalized


class cheb_conv_withSAt(nn.Module):
    '''
    K-order chebyshev graph convolution
    '''

    def __init__(self, K, cheb_polynomials, in_channels, out_channels):
        '''
        :param K: int
        :param in_channles: int, num of channels in the input sequence
        :param out_channels: int, num of channels in the output sequence
        '''
        super(cheb_conv_withSAt, self).__init__()
        self.K = K
        self.cheb_polynomials = cheb_polynomials
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.DEVICE = cheb_polynomials[0].device
        self.Theta = nn.ParameterList([nn.Parameter(torch.FloatTensor(in_channels, out_channels).to(self.DEVICE)) for _ in range(K)])

    def forward(self, x, spatial_attention):
        '''
        Chebyshev graph convolution operation
        :param x: (batch_size, N, F_in, T)
        :return: (batch_size, N, F_out, T)
        '''

        batch_size, num_of_vertices, in_channels, num_of_timesteps = x.shape

        outputs = []

        for time_step in range(num_of_timesteps):

            graph_signal = x[:, :, :, time_step]  # (b, N, F_in)

            output = torch.zeros(batch_size, num_of_vertices, self.out_channels).to(self.DEVICE)  # (b, N, F_out)

            for k in range(self.K):

                T_k = self.cheb_polynomials[k]  # (N,N)

                T_k_with_at = T_k.mul(spatial_attention)   # (N,N)*(N,N) = (N,N) 多行和为1, 按着列进行归一化

                theta_k = self.Theta[k]  # (in_channel, out_channel)

                rhs = T_k_with_at.permute(0, 2, 1).matmul(graph_signal)  # (N, N)(b, N, F_in) = (b, N, F_in) 因为是左乘，所以多行和为1变为多列和为1，即一行之和为1，进行左乘

                output = output + rhs.matmul(theta_k)  # (b, N, F_in)(F_in, F_out) = (b, N, F_out)

            outputs.append(output.unsqueeze(-1))  # (b, N, F_out, 1)

        return F.relu(torch.cat(outputs, dim=-1))  # (b, N, F_out, T)


class Temporal_Attention_layer(nn.Module):
    def __init__(self, DEVICE, in_channels, num_of_vertices, num_of_timesteps):
        super(Temporal_Attention_layer, self).__init__()
        self.U1 = nn.Parameter(torch.FloatTensor(num_of_vertices).to(DEVICE))
        self.U2 = nn.Parameter(torch.FloatTensor(in_channels, num_of_vertices).to(DEVICE))
        self.U3 = nn.Parameter(torch.FloatTensor(in_channels).to(DEVICE))
        self.be = nn.Parameter(torch.FloatTensor(1, num_of_timesteps, num_of_timesteps).to(DEVICE))
        self.Ve = nn.Parameter(torch.FloatTensor(num_of_timesteps, num_of_timesteps).to(DEVICE))

    def forward(self, x):
        '''
        :param x: (batch_size, N, F_in, T)
        :return: (B, T, T)
        '''
        _, num_of_vertices, num_of_features, num_of_timesteps = x.shape

        lhs = torch.matmul(torch.matmul(x.permute(0, 3, 2, 1), self.U1), self.U2)
        # x:(B, N, F_in, T) -> (B, T, F_in, N)
        # (B, T, F_in, N)(N) -> (B,T,F_in)
        # (B,T,F_in)(F_in,N)->(B,T,N)

        rhs = torch.matmul(self.U3, x)  # (F)(B,N,F,T)->(B, N, T)

        product = torch.matmul(lhs, rhs)  # (B,T,N)(B,N,T)->(B,T,T)

        E = torch.matmul(self.Ve, torch.sigmoid(product + self.be))  # (B, T, T)

        E_normalized = F.softmax(E, dim=1)

        return E_normalized


class cheb_conv(nn.Module):
    '''
    K-order chebyshev graph convolution
    '''

    def __init__(self, K, cheb_polynomials, in_channels, out_channels):
        '''
        :param K: int
        :param in_channles: int, num of channels in the input sequence
        :param out_channels: int, num of channels in the output sequence
        '''
        super(cheb_conv, self).__init__()
        self.K = K
        self.cheb_polynomials = cheb_polynomials
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.DEVICE = cheb_polynomials[0].device
        self.Theta = nn.ParameterList([nn.Parameter(torch.FloatTensor(in_channels, out_channels).to(self.DEVICE)) for _ in range(K)])

    def forward(self, x):
        '''
        Chebyshev graph convolution operation
        :param x: (batch_size, N, F_in, T)
        :return: (batch_size, N, F_out, T)
        '''

        batch_size, num_of_vertices, in_channels, num_of_timesteps = x.shape

        outputs = []

        for time_step in range(num_of_timesteps):

            graph_signal = x[:, :, :, time_step]  # (b, N, F_in)

            output = torch.zeros(batch_size, num_of_vertices, self.out_channels).to(self.DEVICE)  # (b, N, F_out)

            for k in range(self.K):

                T_k = self.cheb_polynomials[k]  # (N,N)

                theta_k = self.Theta[k]  # (in_channel, out_channel)

                rhs = graph_signal.permute(0, 2, 1).matmul(T_k).permute(0, 2, 1)

                output = output + rhs.matmul(theta_k)

            outputs.append(output.unsqueeze(-1))

        return F.relu(torch.cat(outputs, dim=-1))


class ASTGCN_block(nn.Module):

    def __init__(self, DEVICE, in_channels, K, nb_chev_filter, nb_time_filter, time_strides, cheb_polynomials, num_of_vertices, num_of_timesteps):
        super(ASTGCN_block, self).__init__()
        self.TAt = Temporal_Attention_layer(DEVICE, in_channels, num_of_vertices, num_of_timesteps)
        self.SAt = Spatial_Attention_layer(DEVICE, in_channels, num_of_vertices, num_of_timesteps)
        self.cheb_conv_SAt = cheb_conv_withSAt(K, cheb_polynomials, in_channels, nb_chev_filter)
        self.time_conv = nn.Conv2d(nb_chev_filter, nb_time_filter, kernel_size=(1, 3), stride=(1, time_strides), padding=(0, 1))
        self.residual_conv = nn.Conv2d(in_channels, nb_time_filter, kernel_size=(1, 1), stride=(1, time_strides))
        self.ln = nn.LayerNorm(nb_time_filter)  #需要将channel放到最后一个维度上

    def forward(self, x):
        '''
        :param x: (batch_size, N, F_in, T)
        :return: (batch_size, N, nb_time_filter, T)
        '''
        batch_size, num_of_vertices, num_of_features, num_of_timesteps = x.shape

        # TAt
        temporal_At = self.TAt(x)  # (b, T, T)

        x_TAt = torch.matmul(x.reshape(batch_size, -1, num_of_timesteps), temporal_At).reshape(batch_size, num_of_vertices, num_of_features, num_of_timesteps)

        # SAt
        spatial_At = self.SAt(x_TAt)

        # cheb gcn
        spatial_gcn = self.cheb_conv_SAt(x, spatial_At)  # (b,N,F,T)
        # spatial_gcn = self.cheb_conv(x)

        # convolution along the time axis
        time_conv_output = self.time_conv(spatial_gcn.permute(0, 2, 1, 3))  # (b,N,F,T)->(b,F,N,T) 用(1,3)的卷积核去做->(b,F,N,T)

        # residual shortcut
        x_residual = self.residual_conv(x.permute(0, 2, 1, 3))  # (b,N,F,T)->(b,F,N,T) 用(1,1)的卷积核去做->(b,F,N,T)

        x_residual = self.ln(F.relu(x_residual + time_conv_output).permute(0, 3, 2, 1)).permute(0, 2, 3, 1)
        # (b,F,N,T)->(b,T,N,F) -ln-> (b,T,N,F)->(b,N,F,T)

        return x_residual


class ASTGCN_submodule(nn.Module):
    """

    References:
        - `Attention based spatial-temporal graph convolutional networks for traffic flow forecasting..
          <https://ojs.aaai.org/index.php/AAAI/article/view/3881>`_.
        - `A PyTorch implementation of the ASTGCN model  (guoshnBJTU)
          <https://github.com/guoshnBJTU/ASTGCN-r-pytorch>`_.

    Args:
        DEVICE(torch.device): Which device use to train.
        num_blocks(int): Number of blocks.
        in_channels(int): Number of input channels.
        K(int): Order of chebyshev polynomial.
        num_chev_filter(int): Number of chebyshev filter.
        num_time_filter(int): Number of time filter.
        time_strides(int): Number of time strides.
        cheb_polynomials(int): Chebyshev Polynomials.
        pred_step(int): Number of steps of prediction.
        len_input(int): Number of steps of sequence input.
        num_node(int): Number of nodes.
    """

    def __init__(self, DEVICE, num_blocks, in_channels, K, num_chev_filter, num_time_filter, time_strides, cheb_polynomials, pred_step, len_input, num_node):

        super(ASTGCN_submodule, self).__init__()

        self.BlockList = nn.ModuleList([ASTGCN_block(DEVICE, in_channels, K, num_chev_filter, num_time_filter, time_strides, cheb_polynomials, num_node, len_input)])

        self.BlockList.extend([ASTGCN_block(DEVICE, num_time_filter, K, num_chev_filter, num_time_filter, 1, cheb_polynomials, num_node, len_input//time_strides) for _ in range(num_blocks-1)])

        self.final_conv = nn.Conv2d(int(len_input/time_strides), pred_step, kernel_size=(1, num_time_filter))

        self.DEVICE = DEVICE

        self.to(DEVICE)

    def forward(self, x):
        '''
        :param x: (B, N_nodes, F_in, T_in)
        :return: (B, N_nodes, T_out)
        '''
        for block in self.BlockList:
            x = block(x)

        output = self.final_conv(x.permute(0, 3, 1, 2))[:, :, :, -1].permute(0, 2, 1)
        # (b,N,F,T)->(b,T,N,F)-conv<1,F>->(b,c_out*T,N,1)->(b,c_out*T,N)->(b,N,T)

        return output


def make_model(DEVICE, nb_block, in_channels, K, nb_chev_filter, nb_time_filter, time_strides, L_tilde, num_for_predict, len_input, num_of_vertices):
    '''

    :param DEVICE:
    :param nb_block:
    :param in_channels:
    :param K:
    :param nb_chev_filter:
    :param nb_time_filter:
    :param time_strides:
    :param cheb_polynomials:
    :param nb_predict_step:
    :param len_input
    :return:
    '''
    cheb_polynomials = [torch.from_numpy(i).type(torch.FloatTensor).to(DEVICE) for i in cheb_polynomial(L_tilde, K)]
    model = ASTGCN_submodule(DEVICE, nb_block, in_channels, K, nb_chev_filter, nb_time_filter, time_strides, cheb_polynomials, num_for_predict, len_input, num_of_vertices)

    for p in model.parameters():
        if p.dim() > 1:
            nn.init.xavier_uniform_(p)
        else:
            nn.init.uniform_(p)

    return model



def cheb_polynomial(L_tilde, K):
    '''
    compute a list of chebyshev polynomials from T_0 to T_{K-1}

    Parameters
    ----------
    L_tilde: scaled Laplacian, np.ndarray, shape (N, N)

    K: the maximum order of chebyshev polynomials

    Returns
    ----------
    cheb_polynomials: list(np.ndarray), length: K, from T_0 to T_{K-1}

    '''

    N = L_tilde.shape[0]

    cheb_polynomials = [np.identity(N), L_tilde.copy()]

    for i in range(2, K):
        cheb_polynomials.append(2 * L_tilde * cheb_polynomials[i - 1] - cheb_polynomials[i - 2])

    return cheb_polynomials






import tensorflow as tf

from ..model_unit import BaseModel


class ST_ResNet(BaseModel):
    """ST-ResNet is a deep-learning model with an end-to-end structure
    based on unique properties of spatio-temporal data making use of convolution and residual units.

    References:
        - `Deep Spatio-Temporal Residual Networks for Citywide Crowd Flows Prediction (Junbo Zhang et al., 2016)
          <https://arxiv.org/pdf/1610.00081.pdf>`_.
        - `Github repository (lucktroy)
          <https://github.com/lucktroy/DeepST/tree/master/scripts/papers/AAAI17>`_.
    Args:
        width (int): The width of grid data.
        height (int): The height of grid data.
        externai_dim (int): Number of dimensions of external data.
        closeness_len (int): The length of closeness data history. The former consecutive ``closeness_len`` time slots
            of data will be used as closeness history.
        period_len (int): The length of period data history. The data of exact same time slots in former consecutive
            ``period_len`` days will be used as period history.
        trend_len (int): The length of trend data history. The data of exact same time slots in former consecutive
            ``trend_len`` weeks (every seven days) will be used as trend history.
        num_residual_unit (int): Number of residual units. Default: 4
        kernel_size (int): Kernel size in Convolutional neural networks. Default: 3
        lr (float): Learning rate. Default: 1e-5
        code_version (str): Current version of this model code.
        model_dir (str): The directory to store model files. Default:'model_dir'
        conv_filters (int):  the Number of filters in the convolution. Default: 64
        gpu_device (str): To specify the GPU to use. Default: '0'
    """
    def __init__(self,
                 width,
                 height,
                 external_dim,
                 closeness_len,
                 period_len,
                 trend_len,
                 num_residual_unit=4,
                 kernel_size=3,
                 lr=5e-5,
                 model_dir='model_dir',
                 code_version='QuickStart',
                 conv_filters=64,
                 gpu_device='0'):

        super(ST_ResNet, self).__init__(code_version=code_version, model_dir=model_dir, gpu_device=gpu_device)
        
        self._width = width
        self._height = height
        self._closeness_len = closeness_len
        self._period_len = period_len
        self._trend_len = trend_len
        self._conv_filters = conv_filters
        self._kernel_size = kernel_size
        self._external_dim = external_dim
        self._num_residual_unit = num_residual_unit
        self._lr = lr

    def build(self):
        with self._graph.as_default():

            target_conf = []

            if self._closeness_len is not None and self._closeness_len > 0:
                c_conf = tf.placeholder(tf.float32, [None, self._height, self._width, self._closeness_len], name='c')
                self._input['closeness_feature'] = c_conf.name
                target_conf.append(c_conf)

            if self._period_len is not None and self._period_len > 0:
                p_conf = tf.placeholder(tf.float32, [None, self._height, self._width, self._period_len], name='p')
                self._input['period_feature'] = p_conf.name
                target_conf.append(p_conf)

            if self._trend_len is not None and self._trend_len > 0:
                t_conf = tf.placeholder(tf.float32, [None, self._height, self._width, self._trend_len], name='t')
                self._input['trend_feature'] = t_conf.name
                target_conf.append(t_conf)

            target = tf.placeholder(tf.float32, [None, self._height, self._width, 1], name='target')

            self._input['target'] = target.name

            outputs = []
            for conf in target_conf:

                residual_input = tf.layers.conv2d(conf, filters=self._conv_filters,
                                                  kernel_size=[self._kernel_size, self._kernel_size],
                                                  padding='SAME', activation=tf.nn.relu)

                def residual_unit(x):
                    residual_output = tf.nn.relu(x)
                    residual_output = tf.layers.conv2d(residual_output, filters=self._conv_filters,
                                                       kernel_size=[self._kernel_size, self._kernel_size], padding='SAME')
                    residual_output = tf.nn.relu(residual_output)
                    residual_output = tf.layers.conv2d(residual_output, filters=self._conv_filters,
                                                       kernel_size=[self._kernel_size, self._kernel_size], padding='SAME')
                    return residual_output + x

                for i in range(self._num_residual_unit):
                    residual_input = residual_unit(residual_input)

                outputs.append(tf.layers.conv2d(tf.nn.relu(residual_input), filters=self._conv_filters,
                                                kernel_size=[self._kernel_size, self._kernel_size], padding='SAME'))

            if len(outputs) == 1:
                x = outputs[0]
            else:
                fusion_weight = tf.Variable(tf.random_normal([len(outputs), ]))
                for i in range(len(outputs)):
                    outputs[i] = fusion_weight[i] * outputs[i]
                x = tf.reduce_sum(outputs, axis=0)

            # external dims
            if self._external_dim is not None and self._external_dim > 0:
                external_input = tf.placeholder(tf.float32, [None, self._external_dim])
                self._input['external_feature'] = external_input.name
                external_dense = tf.layers.dense(inputs=external_input, units=10)
                external_dense = tf.tile(tf.reshape(external_dense, [-1, 1, 1, 10]),
                                         [1, self._height, self._width, 1])
                x = tf.concat([x, external_dense], axis=-1)

            x = tf.layers.dense(x, units=1, name='prediction', activation=tf.nn.sigmoid)

            loss = tf.sqrt(tf.reduce_mean(tf.square(x - target)), name='loss')
            train_op = tf.train.AdamOptimizer(self._lr).minimize(loss)

            self._output['prediction'] = x.name
            self._output['loss'] = loss.name

            self._op['train_op'] = train_op.name

        super(ST_ResNet, self).build()

    def _get_feed_dict(self, closeness_feature=None, period_feature=None, trend_feature=None,
                       target=None, external_feature=None):
        '''
        The method to get feet dict for tensorflow model.

        Users may modify this method according to the format of input.

        Args:
            closeness_feature (np.ndarray or ``None``): The closeness history data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, closeness_len].
            period_feature (np.ndarray or ``None``): The period history data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, period_len].
            trend_feature (np.ndarray or ``None``): The trend history data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, trend_len].
            target (np.ndarray or ``None``): The target value data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, 1].
            external_feature (np.ndarray or ``None``): The external feature data.
                If type is np.ndaaray, its shape is [time_slot_num, feature_num].
        '''
        feed_dict = {}
        if target is not None:
            feed_dict['target'] = target
        if self._external_dim is not None and self._external_dim > 0:
            feed_dict['external_feature'] = external_feature
        if self._closeness_len is not None and self._closeness_len > 0:
            feed_dict['closeness_feature'] = closeness_feature
        if self._period_len is not None and self._period_len > 0:
            feed_dict['period_feature'] = period_feature
        if self._trend_len is not None and self._trend_len > 0:
            feed_dict['trend_feature'] = trend_feature
        return feed_dict



import tensorflow as tf

from ..model_unit import BaseModel
from ..model_unit import DCGRUCell

from tensorflow.contrib import legacy_seq2seq


class DCRNN(BaseModel):
    """

    References:
        - `Diffusion convolutional recurrent neural network: Data-driven traffic forecasting (Li Yaguang, et al., 2017)
          <https://arxiv.org/pdf/1707.01926.pdf>`_.
        - `A TensorFlow implementation of Diffusion Convolutional Recurrent Neural Network (liyaguang)
          <https://github.com/liyaguang/DCRNN>`_.

    Args:
        num_node(int): Number of nodes in the graph, e.g. number of stations in NYC-Bike dataset.
        num_diffusion_matrix: Number of diffusion matrix used in model.
        num_rnn_units: Number of RNN units.
        num_rnn_layers: Number of RNN layers
        max_diffusion_step: Number of diffusion steps
        seq_len: Input sequence length
        use_curriculum_learning(bool): model's prediction (True) or the previous ground truth in training (False).
        input_dim: Dimension of the input feature
        output_dim: Dimension of the output feature
        cl_decay_steps: When use_curriculum_learning=True, cl_decay_steps is used to adjust the ratio of using ground
            true labels, where with more training steps, the ratio drops.
        target_len(int): Output sequence length.
        lr(float): Learning rate
        epsilon: epsilon in Adam
        optimizer_name(str): 'sgd' or 'Adam' optimizer
        code_version(str): Current version of this model code, which will be used as filename for saving the model
        model_dir(str): The directory to store model files. Default:'model_dir'.
        gpu_device(str): To specify the GPU to use. Default: '0'.
    """
    def __init__(self,
                 num_node,
                 num_diffusion_matrix,
                 num_rnn_units=64,
                 num_rnn_layers=1,
                 max_diffusion_step=2,
                 seq_len=6,
                 use_curriculum_learning=False,
                 input_dim=1,
                 output_dim=1,
                 cl_decay_steps=1000,
                 target_len=1,
                 lr=1e-4,
                 epsilon=1e-3,
                 optimizer_name='Adam',
                 code_version='DCRNN-QuickStart',
                 model_dir='model_dir',
                 gpu_device='0', **kwargs):

        super(DCRNN, self).__init__(code_version=code_version, model_dir=model_dir, gpu_device=gpu_device)

        self._num_node = num_node
        self._num_diffusion_matrix = num_diffusion_matrix
        self._num_rnn_units = num_rnn_units
        self._num_rnn_layers = num_rnn_layers
        self._max_diffusion_step = max_diffusion_step
        self._seq_len = seq_len
        self._use_curriculum_learning = use_curriculum_learning
        self._input_dim = input_dim
        self._output_dim = output_dim
        self._target_len = target_len
        self._cl_decay_steps = cl_decay_steps
        self._optimizer_name = optimizer_name
        self._lr = lr
        # self._batch_size = batch_size
        self._epsilon = epsilon

    def build(self, init_vars=True, max_to_keep=5):
        with self._graph.as_default():
            inputs = tf.placeholder(tf.float32, shape=(None, self._seq_len,
                                                       self._num_node, self._input_dim), name='inputs')
            labels = tf.placeholder(tf.float32, shape=(None, self._target_len,
                                                       self._num_node, self._output_dim), name='labels')

            diffusion_matrix = tf.placeholder(tf.float32, shape=(self._num_diffusion_matrix, self._num_node,
                                                                 self._num_node), name='diffusion_matrix')

            batch_size = tf.shape(inputs)[0]

            self._input['inputs'] = inputs.name
            self._input['target'] = labels.name
            self._input['diffusion_matrix'] = diffusion_matrix.name

            go_symbol = tf.zeros(shape=(tf.shape(inputs)[0], self._num_node * self._output_dim))

            cell = DCGRUCell(self._num_rnn_units, self._input_dim, self._num_diffusion_matrix, diffusion_matrix,
                             max_diffusion_step=self._max_diffusion_step, num_node=self._num_node)

            cell_with_projection = DCGRUCell(self._num_rnn_units, self._input_dim,
                                             self._num_diffusion_matrix, diffusion_matrix,
                                             max_diffusion_step=self._max_diffusion_step,
                                             num_node=self._num_node, num_proj=self._output_dim)

            encoding_cells = [cell] * self._num_rnn_layers
            decoding_cells = [cell] * (self._num_rnn_layers - 1) + [cell_with_projection]
            encoding_cells = tf.contrib.rnn.MultiRNNCell(encoding_cells, state_is_tuple=True)
            decoding_cells = tf.contrib.rnn.MultiRNNCell(decoding_cells, state_is_tuple=True)

            global_step = tf.train.get_or_create_global_step()

            # Outputs: (batch_size, timesteps, num_node, output_dim)
            with tf.variable_scope('DCRNN_SEQ'):
                inputs_unstack = tf.unstack(tf.reshape(inputs, (batch_size,
                                                                self._seq_len, self._num_node * self._input_dim)),
                                            axis=1)
                labels_unstack = tf.unstack(
                    tf.reshape(labels[..., :self._output_dim],
                               (batch_size, self._target_len, self._num_node * self._output_dim)), axis=1)
                labels_unstack.insert(0, go_symbol)

                def _compute_sampling_threshold(global_step, k):
                    """
                    Computes the sampling probability for scheduled sampling using inverse sigmoid.
                    global_step:
                    k:
                    :return:
                    """
                    return tf.cast(k / (k + tf.exp(global_step / k)), tf.float32)

                def _loop_function_train(prev, i):
                    # Return either the model's prediction or the previous ground truth in training.
                    if self._use_curriculum_learning:
                        c = tf.random_uniform((), minval=0, maxval=1.)
                        threshold = _compute_sampling_threshold(global_step, self._cl_decay_steps)
                        result = tf.cond(tf.less(c, threshold), lambda: labels_unstack[i], lambda: prev)
                    else:
                        result = labels_unstack[i]
                    return result

                def _loop_function_test(prev, i):
                    # Return the prediction of the model in testing.
                    return prev

                a, enc_state = tf.contrib.rnn.static_rnn(encoding_cells, inputs_unstack, dtype=tf.float32)

                with tf.variable_scope('train', reuse=False):
                    train_outputs, _ = legacy_seq2seq.rnn_decoder(labels_unstack, enc_state, decoding_cells,
                                                                  loop_function=_loop_function_train)
                with tf.variable_scope('text', reuse=True):
                    test_outputs, _ = legacy_seq2seq.rnn_decoder(labels_unstack, enc_state, decoding_cells,
                                                                 loop_function=_loop_function_test)

            # Project the output to output_dim.
            train_outputs = tf.stack(train_outputs[:-1], axis=1)
            test_outputs = tf.stack(test_outputs[:-1], axis=1)

            # Configure optimizer
            optimizer = tf.train.AdamOptimizer(self._lr, epsilon=float(self._epsilon))

            if self._optimizer_name == 'sgd':
                optimizer = tf.train.GradientDescentOptimizer(self._lr)

            loss = tf.sqrt(tf.reduce_mean(tf.square(train_outputs - labels[:, :, :, 0])))

            train_op = optimizer.minimize(loss)

            self._output['prediction'] = test_outputs.name
            self._output['loss'] = loss.name
            self._op['train_op'] = train_op.name

        super(DCRNN, self).build(init_vars=init_vars, max_to_keep=5)

    # Define your '_get_feed_dict function‘, map your input to the tf-model
    def _get_feed_dict(self,
                       inputs,
                       diffusion_matrix,
                       target=None,):
        feed_dict = {
            'inputs': inputs,
            'diffusion_matrix': diffusion_matrix,
        }
        if target is not None:
            feed_dict['target'] = target
        return feed_dict



import tensorflow as tf
from tensorflow.contrib.framework import nest
from ..model_unit import BaseModel


class GeoMAN(BaseModel):
    """Multi-level Attention Networks for Geo-sensory Time Series Prediction (GeoMAN)

            GeoMAN consists of two major parts: 1) A multi-level attention mechanism (including both local and global
            spatial attentions in encoder and temporal attention in decoder) to model the dynamic spatio-temporal
            dependencies; 2) A general fusion module to incorporate the external factors from different domains (e.g.,
            meteorology, time of day and land use).

            References:
                - `GeoMAN: Multi-level Attention Networks for Geo-sensory Time Series Prediction (Liang Yuxuan, et al., 2018)
                  <https://www.ijcai.org/proceedings/2018/0476.pdf>`_.
                - `An easy implement of GeoMAN using TensorFlow (yoshall & CastleLiang)
                  <https://github.com/yoshall/GeoMAN>`_.

            Args:
                total_sensers (int): The number of total sensors used in global attention mechanism.
                input_dim (int): The number of dimensions of the target sensor's input.
                external_dim (int): The number of dimensions of the external features.
                output_dim (int): The number of dimensions of the target sensor's output.
                input_steps (int): The length of historical input data, a.k.a, input timesteps.
                output_steps (int): The number of steps that need prediction by one piece of history data, a.k.a, output
                    timesteps. Have to be 1 now.
                n_stacked_layers (int): The number of LSTM layers stacked in both encoder and decoder (These two are the
                    same). Default: 2
                n_encoder_hidden_units (int): The number of hidden units in each layer of encoder. Default: 128
                n_decoder_hidden_units (int): The number of hidden units in each layer of decoder. Default: 128
                dropout_rate (float): Dropout rate of LSTM layers in both encoder and decoder. Default: 0.3
                lr (float): Learning rate. Default: 0.001
                gc_rate (float): A clipping ratio for all the gradients. This operation normalizes all gradients so that
                    their L2-norms are less than or equal to ``gc_rate``. Default: 2.5
                code_version (str): Current version of this model code. Default: 'GeoMAN-QuickStart'
                model_dir (str): The directory to store model files. Default:'model_dir'
                gpu_device (str): To specify the GPU to use. Default: '0'
                **kwargs (dict): Reserved for future use. May be used to pass parameters to class ``BaseModel``.
            """
    def __init__(self,
                 total_sensers,
                 input_dim,
                 external_dim,
                 output_dim,
                 input_steps,
                 output_steps,
                 n_stacked_layers=2,
                 n_encoder_hidden_units=128,
                 n_decoder_hidden_units=128,
                 dropout_rate=0.3,
                 lr=0.001,
                 gc_rate=2.5,
                 code_version='GeoMAN-QuickStart',
                 model_dir='model_dir',
                 gpu_device='0',
                 **kwargs):

        super(GeoMAN, self).__init__(code_version=code_version, model_dir=model_dir, gpu_device=gpu_device)

        # Architecture
        self._n_stacked_layers = n_stacked_layers
        self._n_encoder_hidden_units = n_encoder_hidden_units
        self._n_decoder_hidden_units = n_decoder_hidden_units
        self._n_output_decoder = output_dim  # n_output_decoder

        self._n_steps_encoder = input_steps  # encoder_steps
        self._n_steps_decoder = output_steps  # decoder_steps
        self._n_input_encoder = input_dim  # n_input_encoder
        self._n_sensers = total_sensers  # n_sensers
        self._n_external_input = external_dim  # external_dim

        # Hyperparameters
        self._dropout_rate = dropout_rate
        self._lr = lr
        self._gc_rate = gc_rate

    def build(self, init_vars=True, max_to_keep=5):
        with self._graph.as_default():
            with tf.variable_scope('inputs'):
                local_features = tf.placeholder(tf.float32, shape=[None, self._n_steps_encoder, self._n_input_encoder],
                                                name='local_features')
                global_features = tf.placeholder(tf.float32, shape=[None, self._n_steps_encoder, self._n_sensers],
                                                 name='global_features')
                external_features = tf.placeholder(tf.float32,
                                                   shape=[None, self._n_steps_decoder, self._n_external_input],
                                                   name='external_features')
                local_attn_states = tf.placeholder(tf.float32,
                                                   shape=[None, self._n_input_encoder, self._n_steps_encoder],
                                                   name='local_attn_states')
                global_attn_states = tf.placeholder(tf.float32, shape=[None, self._n_sensers, self._n_input_encoder,
                                                                       self._n_steps_encoder],
                                                    name='global_attn_states')
            with tf.variable_scope('ground_truth'):
                targets = tf.placeholder(tf.float32, [None, self._n_steps_decoder, self._n_output_decoder])

            self._input['local_features'] = local_features.name
            self._input['global_features'] = global_features.name
            self._input['external_features'] = external_features.name
            self._input['local_attn_states'] = local_attn_states.name
            self._input['global_attn_states'] = global_attn_states.name
            self._input['targets'] = targets.name

            predict_layer = tf.keras.layers.Dense(units=self._n_output_decoder,
                                                  kernel_initializer=tf.truncated_normal_initializer,
                                                  bias_initializer=tf.constant_initializer(0.),
                                                  use_bias=True)

            def _build_cells(n_hidden_units):
                cells = []
                for i in range(self._n_stacked_layers):
                    with tf.variable_scope(f'LSTM_{i}'):
                        cell = tf.contrib.rnn.BasicLSTMCell(n_hidden_units,
                                                            forget_bias=1.0,
                                                            state_is_tuple=True)
                        cell = tf.nn.rnn_cell.DropoutWrapper(cell, output_keep_prob=1.0 - self._dropout_rate)
                        cells.append(cell)
                encoder_cell = tf.contrib.rnn.MultiRNNCell(cells)
                return encoder_cell

            def _loop_function(prev):
                """loop function used in the decoder to generate the next input"""
                return predict_layer(prev)

            def _get_MSE_loss(y_true, y_pred):
                return tf.reduce_mean(tf.pow(y_true - y_pred, 2), name='MSE_loss')

            def _get_l2reg_loss():
                # l2 loss
                reg_loss = 0
                for tf_var in tf.trainable_variables():
                    if 'kernel:' in tf_var.name or 'bias:' in tf_var.name:
                        reg_loss += tf.reduce_mean(tf.nn.l2_loss(tf_var))
                return 0.001 * reg_loss

            def _spatial_attention(local_features,  # x and X
                                   global_features,
                                   local_attention_states,
                                   global_attention_states,
                                   encoder_cells,  # to acquire h_{t-1}, s_{t-1}
                                   ):
                batch_size = tf.shape(local_features[0])[0]
                output_size = encoder_cells.output_size
                with tf.variable_scope('spatial_attention'):
                    with tf.variable_scope('local_spatial_attn'):
                        local_attn_length = local_attention_states.get_shape()[1].value  # n_input_encoder
                        local_attn_size = local_attention_states.get_shape()[2].value  # n_steps_encoder
                        local_attn = tf.zeros([batch_size, local_attn_length])

                        #  Add local features in attention
                        x_ik = tf.reshape(local_attention_states,
                                          [-1, local_attn_length, 1, local_attn_size])  # features
                        Ul = tf.get_variable('spati_atten_Ul', [1, 1, local_attn_size, local_attn_size])
                        Ul_x = tf.nn.conv2d(x_ik, Ul, [1, 1, 1, 1], 'SAME')  # U_l * x^{i,k}
                        vl = tf.get_variable('spati_atten_vl', [local_attn_size])  # v_l

                        def _local_spatial_attention(query):
                            # If the query is a tuple (when stacked RNN/LSTM), flatten it
                            if hasattr(query, "__iter__"):
                                query_list = nest.flatten(query)
                                for q in query_list:
                                    ndims = q.get_shape().ndims
                                    if ndims:
                                        assert ndims == 2
                                query = tf.concat(query_list, 1)
                            with tf.variable_scope('local_spatial_attn_Wl'):
                                h_s = query
                                Wl_hs_bl = tf.keras.layers.Dense(units=local_attn_size, use_bias=True)(h_s)
                                Wl_hs_bl = tf.reshape(Wl_hs_bl, [-1, 1, 1, local_attn_size])
                                score = tf.reduce_sum(vl * tf.nn.tanh(Wl_hs_bl + Ul_x),
                                                      [2, 3])  # ! Ux is a 4 dims matrix, have to use reduce_sum here
                                attention_weights = tf.nn.softmax(score)
                            return attention_weights

                    with tf.variable_scope('global_spatial_attn'):
                        global_attn_length = global_attention_states.get_shape()[1].value  # n_sensor
                        global_n_input = global_attention_states.get_shape()[2].value  # n_input_dim
                        global_attn_size = global_attention_states.get_shape()[3].value  # n_input_dim
                        global_attn = tf.zeros([batch_size, global_attn_length])

                        # Add global features in attention
                        Xl = tf.reshape(global_attention_states,
                                        [-1, global_attn_length, global_n_input, global_attn_size])
                        Wg_ug = tf.get_variable('spati_atten_Wg_ug',
                                                [1, global_n_input, global_attn_size, global_attn_size])
                        Wg_Xl_ug = tf.nn.conv2d(Xl, Wg_ug, [1, 1, 1, 1], 'SAME')
                        vg = tf.get_variable('spati_atten_vg', [local_attn_size])

                        # TODO: add U_g * y^l here, where y^l is the first column of local inputs.

                        def _global_spatial_attention(query):
                            if hasattr(query, "__iter__"):
                                query_list = nest.flatten(query)
                                for q in query_list:  # Check that ndims == 2 if specified.
                                    ndims = q.get_shape().ndims
                                    if ndims:
                                        assert ndims == 2
                                query = tf.concat(query_list, 1)
                            with tf.variable_scope('global_spatial_attn_Wl'):
                                h_s = query
                                Wg_hs_bg = tf.keras.layers.Dense(units=global_attn_size, use_bias=True)(h_s)
                                Wg_hs_bg = tf.reshape(Wg_hs_bg, [-1, 1, 1, global_attn_size])
                                score = tf.reduce_sum(vg * tf.nn.tanh(Wg_hs_bg + Wg_Xl_ug), [2, 3])
                                attention_weights = tf.nn.softmax(score)
                                # Sometimes it's not easy to find a measurement to denote similarity between sensors,
                                # here we omit such prior knowledge in eq.[4].
                                # You can use "a = nn_ops.softmax((1-lambda)*s + lambda*sim)" to encode similarity info,
                                # where:
                                #     sim: a vector with length n_sensors, describing the sim between the target sensor and the others
                                #     lambda: a trade-off.
                                # attention_weights = tf.softmax((1-self.sm_rate)*score+self.sm_rate*self.similarity_graph)
                            return attention_weights

                    # Init
                    zeros = [tf.zeros([batch_size, output_size]) for i in range(2)]
                    initial_state = [zeros for _ in range(len(encoder_cells._cells))]
                    state = initial_state

                    # For each timestep
                    outputs = []
                    attn_weights = []
                    for i, (local_input, global_input) in enumerate(zip(local_features, global_features)):
                        if i > 0: tf.get_variable_scope().reuse_variables()

                        local_context_vector = local_attn * local_input
                        global_context_vector = global_attn * global_input
                        x_t = tf.concat([local_context_vector, global_context_vector], axis=1)
                        encoder_output, state = encoder_cells(x_t, state)  # Update states

                        with tf.variable_scope('local_spatial_attn'):
                            local_attn = _local_spatial_attention(state)
                        with tf.variable_scope('global_spatial_attn'):
                            global_attn = _global_spatial_attention(state)
                        attn_weights.append((local_attn, global_attn))
                        outputs.append(encoder_output)

                return outputs, state, attn_weights

            def _temporal_attention(decoder_inputs,
                                    external_features,
                                    inital_states,  # the first time, the output of encoder
                                    attention_states,  # h_o
                                    decoder_cells):
                batch_size = tf.shape(decoder_inputs[0])[0]
                output_size = decoder_cells.output_size
                input_size = decoder_inputs[0].get_shape().with_rank(2)[1]  # ?
                state = inital_states
                with tf.variable_scope('temperal_attention'):
                    attn_length = attention_states.get_shape()[1].value
                    attn_size = attention_states.get_shape()[2].value

                    h_o = tf.reshape(attention_states, [-1, attn_length, 1, attn_size])
                    W_d = tf.get_variable('temperal_attn_Wd', [1, 1, attn_size, attn_size])
                    W_h = tf.nn.conv2d(h_o, W_d, [1, 1, 1, 1], 'SAME')
                    v_d = tf.get_variable('temperal_attn_vd', [attn_size])

                    def _attention(query):
                        if hasattr(query, "__iter__"):
                            query_list = nest.flatten(query)
                            for q in query_list:  # Check that ndims == 2 if specified.
                                ndims = q.get_shape().ndims
                                if ndims:
                                    assert ndims == 2
                            query = tf.concat(query_list, 1)
                        with tf.variable_scope('attention'):
                            d_s = query
                            W_ds_b = tf.keras.layers.Dense(units=attn_size, use_bias=True)(d_s)
                            W_ds_b = tf.reshape(W_ds_b, [-1, 1, 1, attn_size])
                            score = tf.reduce_sum(v_d * tf.nn.tanh(W_ds_b + W_h), [2, 3])
                            attention_weights = tf.nn.softmax(score)
                            context_vector = tf.reduce_sum(
                                tf.reshape(attention_weights, [-1, attn_length, 1, 1]) * h_o, [1, 2])
                            context_vector = tf.reshape(context_vector, [-1, attn_size])
                        return context_vector

                    # Init
                    inital_attn = tf.zeros([batch_size, output_size])
                    attn = inital_attn
                    outputs = []

                    prev_decoder_output = None  # d_{t-1}
                    for i, (decoder_input, external_input) in enumerate(zip(decoder_inputs, external_features)):
                        if i > 0: tf.get_variable_scope().reuse_variables()
                        if prev_decoder_output is not None and _loop_function is not None:
                            with tf.variable_scope('loop_function', reuse=True):
                                decoder_input = _loop_function(prev_decoder_output)
                        x = tf.concat([decoder_input, external_input, attn], axis=1)
                        x = tf.keras.layers.Dense(units=input_size, use_bias=True)(x)
                        decoder_output, state = decoder_cells(x, state)
                        # Update attention weights
                        attn = _attention(state)
                        # Attention output projection
                        with tf.variable_scope("attn_output_projection"):
                            x = tf.concat([decoder_output, attn], axis=1)
                            output = tf.keras.layers.Dense(units=output_size, use_bias=True)(x)
                        outputs.append(output)
                        prev_decoder_output = output
                return outputs, state

            # Handle data
            local_features, global_features, external_features, targets, decoder_inputs = input_transform(
                local_features, global_features, external_features, targets)

            with tf.variable_scope('GeoMAN'):
                with tf.variable_scope('encoder'):
                    encoder_cells = _build_cells(self._n_encoder_hidden_units)
                    encoder_outputs, encoder_state, attn_weights = _spatial_attention(local_features,
                                                                                      global_features,
                                                                                      local_attn_states,
                                                                                      global_attn_states,
                                                                                      encoder_cells)
                    top_states = [tf.reshape(e, [-1, 1, encoder_cells.output_size]) for e in encoder_outputs]
                    attention_states = tf.concat(top_states, 1)

                with tf.variable_scope('decoder'):
                    decoder_cells = _build_cells(self._n_decoder_hidden_units)
                    decoder_outputs, states = _temporal_attention(decoder_inputs,
                                                                  external_features,
                                                                  encoder_state,
                                                                  attention_states,
                                                                  decoder_cells)

                with tf.variable_scope('prediction'):
                    predictions = []
                    for decoder_output in decoder_outputs:
                        predictions.append(predict_layer(decoder_output))
                    predictions = tf.stack(predictions, axis=1, name='predictions')

                with tf.variable_scope('loss'):
                    targets = tf.stack(targets, axis=1, name='targets')
                    loss = tf.add(_get_MSE_loss(targets, predictions), _get_l2reg_loss(), name='loss')

                with tf.variable_scope('train_op'):
                    global_step = tf.train.get_or_create_global_step()
                    optimizer = tf.train.AdamOptimizer(self._lr)
                    gradients, variables = zip(*optimizer.compute_gradients(loss))
                    gradients, _ = tf.clip_by_global_norm(gradients, self._gc_rate)  # clip norm
                    train_op = optimizer.apply_gradients(zip(gradients, variables), global_step)

                # record output
                self._output['prediction'] = predictions.name
                self._output['loss'] = loss.name
                # record op
                self._op['train_op'] = train_op.name

        super(GeoMAN, self).build(init_vars=init_vars, max_to_keep=5)

    def _get_feed_dict(self,
                       local_features,
                       global_features,
                       local_attn_states,
                       global_attn_states,
                       external_features,
                       targets):
        """The method to get feet dict for tensorflow model.

        Users may modify this method according to the format of input.

        Args:
            local_features (np.ndarray): All the time series generated by the target sensor i, including one target
                series and other feature series, with shape `(batch, input_steps, input_dim)`.
            global_features (np.ndarray): Target series generated by all the sensors, with shape `(batch, input_steps,
                total_sensors)`.
            local_attn_states (np.ndarray): Equals to ``local_features`` swapped ``input_steps`` and ``input_dim`` axis,
                with shape `(batch, input_dim, input_steps)`.
            global_attn_states (np.ndarray): All time series generated by all sensors, with shape `(batch,
                total_sensors, input_dim, input_steps)`.
            external_features (np.ndarray): Fused external factors, e.g., temporal factors: meteorology and spatial
                factors: POIs density, with shape `(batch, output_steps, external_dim)`. All features have to be
                time series.
            targets (np.ndarray): Target sensor's labels, with shape `(batch, output_steps, output_dim)`.
        """
        feed_dict = {'local_features': local_features, 'global_features': global_features,
                     'local_attn_states': local_attn_states, 'global_attn_states': global_attn_states,
                     'external_features': external_features, 'targets': targets}
        return feed_dict


def input_transform(local_features,
                    global_features,
                    external_features,
                    targets):
    """Split the model's inputs from matrices to lists on timesteps axis."""
    local_features = split_timesteps(local_features)
    global_features = split_timesteps(global_features)
    external_features = split_timesteps(external_features)
    targets = split_timesteps(targets)
    decoder_inputs = [tf.zeros_like(targets[0], dtype=tf.float32)] + targets[:-1]  # useless when loop func is employed
    return local_features, global_features, external_features, targets, decoder_inputs


def split_timesteps(inputs):
    """Split the input matrix from (batch, timesteps, input_dim) to a step list ([[batch, input_dim], ..., ])."""
    timesteps = inputs.get_shape()[1].value
    feature_dims = inputs.get_shape()[2].value
    inputs = tf.transpose(inputs, [1, 0, 2])
    inputs = tf.reshape(inputs, [-1, feature_dims])
    inputs = tf.split(inputs, timesteps, 0)
    return inputs






import os
import time
import torch
import torch.nn as nn
import numpy as np
import torch.optim as optim
from UCTB.preprocess.preprocessor import SplitData
from UCTB.model.GraphWaveNet import *
from UCTB.train.LossFunction import masked_mape, masked_mae, masked_rmse


class trainer():
    def __init__(self, in_dim, seq_length, num_nodes, nhid , dropout, lrate, wdecay, device, supports, gcn_bool, addaptadj, aptinit):
        self.model = gwnet(device, num_nodes, dropout, supports=supports, gcn_bool=gcn_bool, addaptadj=addaptadj, aptinit=aptinit, in_dim=in_dim, out_dim=seq_length, residual_channels=nhid, dilation_channels=nhid, skip_channels=nhid * 8, end_channels=nhid * 16)
        self.model.to(device)
        self.optimizer = optim.Adam(self.model.parameters(), lr=lrate, weight_decay=wdecay)
        self.loss = masked_mae
        self.clip = 5

    def train(self, input, real_val):
        self.model.train()
        self.optimizer.zero_grad()
        input = nn.functional.pad(input,(1,0,0,0))
        # print("input",input.shape)
        output = self.model(input)
        output = output.transpose(1,3)
        # print("ouput",output.shape)
        #output = [batch_size,12,num_nodes,1]
        real = torch.unsqueeze(real_val,dim=1)
        predict = output

        loss = self.loss(predict, real, 0.0)
        loss.backward()
        if self.clip is not None:
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.clip)
        self.optimizer.step()
        mape = masked_mape(predict,real,0.0).item()
        rmse = masked_rmse(predict,real,0.0).item()
        return loss.item(),mape,rmse

    def eval(self, input, real_val):
        self.model.eval()
        input = nn.functional.pad(input,(1,0,0,0))
        output = self.model(input)
        output = output.transpose(1,3)
        #output = [batch_size,12,num_nodes,1]
        real = torch.unsqueeze(real_val,dim=1)
        predict = output
        loss = self.loss(predict, real, 0.0)
        mape = masked_mape(predict,real,0.0).item()
        rmse = masked_rmse(predict,real,0.0).item()
        return loss.item(),mape,rmse


def Training(args, dataloader,  device, engine):
    print("start training...", flush=True)
    his_loss = []
    val_time = []
    train_time = []
    for i in range(1, args.epochs + 1):
        train_loss = []
        train_mape = []
        train_rmse = []
        t1 = time.time()
        dataloader['train_loader'].shuffle()
        for iter, (x, y) in enumerate(dataloader['train_loader'].get_iterator()):
            trainx = torch.Tensor(x).to(device)
            trainx = trainx.transpose(1, 3)
            trainy = torch.Tensor(y).to(device)
            trainy = trainy.transpose(1, 3)
            metrics = engine.train(trainx, trainy[:, 0, :, :])
            train_loss.append(metrics[0])
            train_mape.append(metrics[1])
            train_rmse.append(metrics[2])
            if iter % args.print_every == 0:
                log = 'Iter: {:03d}, Train Loss: {:.4f}, Train MAPE: {:.4f}, Train RMSE: {:.4f}'
                print(log.format(iter, train_loss[-1], train_mape[-1], train_rmse[-1]), flush=True)
        t2 = time.time()
        train_time.append(t2 - t1)
        # validation
        valid_loss = []
        valid_mape = []
        valid_rmse = []

        s1 = time.time()
        for iter, (x, y) in enumerate(dataloader['val_loader'].get_iterator()):
            testx = torch.Tensor(x).to(device)
            testx = testx.transpose(1, 3)
            testy = torch.Tensor(y).to(device)
            testy = testy.transpose(1, 3)
            metrics = engine.eval(testx, testy[:, 0, :, :])
            valid_loss.append(metrics[0])
            valid_mape.append(metrics[1])
            valid_rmse.append(metrics[2])
        s2 = time.time()
        log = 'Epoch: {:03d}, Inference Time: {:.4f} secs'
        print(log.format(i, (s2 - s1)))
        val_time.append(s2 - s1)
        mtrain_loss = np.mean(train_loss)
        mtrain_mape = np.mean(train_mape)
        mtrain_rmse = np.mean(train_rmse)

        mvalid_loss = np.mean(valid_loss)
        mvalid_mape = np.mean(valid_mape)
        mvalid_rmse = np.mean(valid_rmse)
        his_loss.append(mvalid_loss)

        log = 'Epoch: {:03d}, Train Loss: {:.4f}, Train MAPE: {:.4f}, Train RMSE: {:.4f}, Valid Loss: {:.4f}, Valid MAPE: {:.4f}, Valid RMSE: {:.4f}, Training Time: {:.4f}/epoch'
        print(log.format(i, mtrain_loss, mtrain_mape, mtrain_rmse, mvalid_loss, mvalid_mape, mvalid_rmse, (t2 - t1)),
              flush=True)
        torch.save(engine.model.state_dict(),
                   os.path.join(args.save, "epoch_" + str(i) + "_" + str(round(mvalid_loss, 2)) + ".pth"))
    print("Average Training Time: {:.4f} secs/epoch".format(np.mean(train_time)))
    print("Average Inference Time: {:.4f} secs".format(np.mean(val_time)))

    return np.argmin(his_loss), his_loss[np.argmin(his_loss)]

def Test(args,dataloader,device,engine,epoch_id, loss_id):
    # Test
    engine.model.load_state_dict(torch.load(
        os.path.join(args.save, "epoch_" + str(epoch_id + 1) + "_" + str(round(loss_id, 2)) + ".pth")))

    outputs = []
    realy = torch.Tensor(dataloader['y_test']).to(device)
    print("realy", realy.shape)

    realy = realy.transpose(1, 3)[:, 0, :, :]

    for iter, (x, y) in enumerate(dataloader['test_loader'].get_iterator()):
        testx = torch.Tensor(x).to(device)
        testx = testx.transpose(1, 3)
        with torch.no_grad():
            preds = engine.model(testx).transpose(1, 3)
        outputs.append(preds.squeeze())

    yhat = torch.cat(outputs, dim=0).cpu().numpy()
    yhat = yhat[:realy.size(0), ...]
    realy = realy.cpu().numpy()

    return yhat

class DataLoader(object):
    def __init__(self, xs, ys, batch_size, pad_with_last_sample=True):
        """
        :param xs:
        :param ys:
        :param batch_size:
        :param pad_with_last_sample: pad with the last sample to make number of samples divisible to batch_size.
        """
        self.batch_size = batch_size
        self.current_ind = 0
        if pad_with_last_sample:
            num_padding = (batch_size - (len(xs) % batch_size)) % batch_size
            x_padding = np.repeat(xs[-1:], num_padding, axis=0)
            y_padding = np.repeat(ys[-1:], num_padding, axis=0)
            xs = np.concatenate([xs, x_padding], axis=0)
            ys = np.concatenate([ys, y_padding], axis=0)
        self.size = len(xs)
        self.num_batch = int(self.size // self.batch_size)
        self.xs = xs
        self.ys = ys

    def shuffle(self):
        permutation = np.random.permutation(self.size)
        xs, ys = self.xs[permutation], self.ys[permutation]
        self.xs = xs
        self.ys = ys

    def get_iterator(self):
        self.current_ind = 0

        def _wrapper():
            while self.current_ind < self.num_batch:
                start_ind = self.batch_size * self.current_ind
                end_ind = min(self.size, self.batch_size * (self.current_ind + 1))
                x_i = self.xs[start_ind: end_ind, ...]
                y_i = self.ys[start_ind: end_ind, ...]
                yield (x_i, y_i)
                self.current_ind += 1

        return _wrapper()



def load_dataset(uctb_data_loader, batch_size, valid_batch_size=None, test_batch_size=None):
    # x_train (num_slots, time_steps, num_stations, input_dims)
    # y_train (num_slots, time_steps, num_stations, input_dims)
    data = {}

    # split data
    train_closeness, val_closeness = SplitData.split_data(uctb_data_loader.train_closeness, [0.9, 0.1])
    train_period, val_period = SplitData.split_data(uctb_data_loader.train_period, [0.9, 0.1])
    train_trend, val_trend = SplitData.split_data(uctb_data_loader.train_trend, [0.9, 0.1])
    train_y, val_y = SplitData.split_data(uctb_data_loader.train_y, [0.9, 0.1])

    # train_x = np.concatenate([train_trend, train_period, train_closeness],axis=2).transpose( # [0,3,1,2] [0,2,1,3]
    if uctb_data_loader.period_len > 0 and uctb_data_loader.trend_len > 0:
        data["x_train"] = np.concatenate([train_trend, train_period, train_closeness], axis=2).transpose([0, 3, 1, 2])
        data["x_val"] = np.concatenate([val_trend, val_period, val_closeness], axis=2).transpose([0, 3, 1, 2])
        data["x_test"] = np.concatenate(
            [uctb_data_loader.test_trend, uctb_data_loader.test_period, uctb_data_loader.test_closeness],
            axis=2).transpose([0, 3, 1, 2])
    else:
        data["x_train"] = train_closeness.transpose([0, 3, 1, 2])
        data["x_val"] = val_closeness.transpose([0, 3, 1, 2])
        data["x_test"] = uctb_data_loader.test_closeness.transpose([0, 3, 1, 2])

    data["y_train"] = train_y[:, np.newaxis]
    data["y_val"] = val_y[:, np.newaxis]
    data["y_test"] = uctb_data_loader.test_y[:, np.newaxis]

    print("x_train", data["x_train"].shape)
    print("y_train", data["y_train"].shape)
    print("x_val", data["x_val"].shape)
    print("y_val", data["y_val"].shape)
    print("x_test", data["x_test"].shape)
    print("y_test", data["y_test"].shape)



    data['train_loader'] = DataLoader(data['x_train'], data['y_train'], batch_size)
    data['val_loader'] = DataLoader(data['x_val'], data['y_val'], valid_batch_size)
    data['test_loader'] = DataLoader(data['x_test'], data['y_test'], test_batch_size)
    return data


import numpy as np
import tensorflow as tf
import heapq

from math import radians, cos, sin, asin, sqrt
from scipy.stats import pearsonr

# Graph Attention Layer
class GAL(object):
    '''
    This class provides static methods for adding Graph Attention Layer.
    '''
    @staticmethod
    def attention_merge_weight(inputs, units, num_head, activation=tf.nn.leaky_relu):
        inputs_shape = inputs.get_shape().with_rank(3)

        num_node = inputs_shape[-2].value
        num_feature = inputs_shape[-1].value

        W = tf.Variable(tf.random_normal([num_feature, units * num_head]))

        # linear transform
        l_t = tf.matmul(tf.reshape(inputs, [-1, num_feature]), W)
        l_t = tf.reshape(l_t, [-1, num_node, num_head, units])

        # compute attention
        a = tf.Variable(tf.random_normal([units * 2, num_head]))

        e = []
        for j in range(1, num_node):
            multi_head_result = []
            for k in range(num_head):
                multi_head_result.append(tf.matmul(tf.concat([l_t[:, 0, k, :], l_t[:, j, k, :]], axis=-1), a[:, k:k+1]))
            e.append(tf.reshape(tf.concat(multi_head_result, axis=-1), [-1, num_head, 1]))

        e = activation(tf.reshape(tf.concat(e, axis=-1), [-1, num_head, num_node-1]))

        alpha = tf.reduce_mean(tf.nn.softmax(e, axis=-1), axis=1, keepdims=True)

        return alpha

    @staticmethod
    def add_ga_layer_matrix(inputs, units, num_head, activation=tf.nn.tanh):
        '''
        This method use Multi-head attention technique to add Graph Attention Layer.

        Args:
            input(ndarray): The set of node features data, with shape [batch, num_node, num_featuer].
            unit(int): The number of merge_gal_units used in GAL.
            num_head(int): The number of multi-head used in GAL.
            activation(function): activation function. default:tf.nn.tanh.
        Returns:
            alpha: The weight matrix after softmax function.
            gc_output: The final GAL aggregated feature representation from input feature.
        '''
        inputs_shape = inputs.get_shape().with_rank(3)

        num_node = inputs_shape[-2].value
        num_feature = inputs_shape[-1].value

        W = tf.Variable(tf.random_normal([num_feature, units * num_head], dtype=tf.float32))

        # linear transform
        l_t = tf.matmul(tf.reshape(inputs, [-1, num_feature]), W)
        l_t = tf.reshape(l_t, [-1, num_node, num_head, units])

        a = tf.Variable(tf.random_normal([units * 2, num_head], dtype=tf.float32))

        e_multi_head = []

        for head_index in range(num_head):

            l_t_i = l_t[:, :, head_index, :]
            a_i = a[:, head_index:head_index+1]

            l_t_i_0 = tf.gather(l_t_i, indices=np.array([e for e in range(num_node)] * num_node), axis=1)
            l_t_i_1 = tf.gather(l_t_i, indices=np.array([[e]*num_node for e in range(num_node)]).reshape([-1,]), axis=1)

            tmp_e = tf.matmul(tf.reshape(tf.concat((l_t_i_0, l_t_i_1), axis=-1), [-1, units*2]), a_i)
            tmp_e = tf.nn.softmax(activation(tf.reshape(tmp_e, [-1, 1, num_node, num_node])), axis=-1)

            e_multi_head.append(tmp_e)

        alpha = tf.concat(e_multi_head, axis=1)

        # Averaging
        gc_output = activation(tf.reduce_mean(tf.matmul(alpha, tf.transpose(l_t, [0, 2, 1, 3])), axis=1))

        return alpha, gc_output

    @staticmethod
    def add_residual_ga_layer(inputs, units, num_head, activation=tf.nn.tanh):
        '''
        Call the add_ga_layer_matrix function to build the Graph Attention Layer, 
        and add the residual layer to optimize the deep neural network.
        '''
        _, gc_output = GAL.add_ga_layer_matrix(inputs, units, num_head, activation=activation)

        gc_output_residual = tf.concat([gc_output, inputs], axis=-1)

        return gc_output_residual


# Graph Convolution Layer
class GCL(object):
    '''
    This class provides static methods for adding Graph Convolution Layer.
    '''
    @staticmethod
    def add_gc_layer(inputs,
                     gcn_k,
                     laplacian_matrix,
                     output_size,
                     dtype=tf.float32,
                     use_bias=True,
                     trainable=True,
                     initializer=None,
                     regularizer=None,
                     activation=tf.nn.tanh):
        '''
        Args:
            Input(ndarray): The input features with shape [batch, num_node, num_feature].
            gcn_k(int): The highest order of Chebyshev Polynomial approximation in GCN.
            laplacian_matrix(ndarray): Laplacian matrix used in GCN, with shape [num_node, num_node].
            output_size(int): Number of output channels.
            dtype: Data type. default:tf.float32.
            use_bias(bool): It determines whether to add bias in the output. default:True.
            trainable(bool): It determines whether `weights` tensor can be trained. default:True.
            initializer: It determines whether the "weight" tensor is initialized. default:None.
            regularizer: It determines whether the "weight" tensor is regularized. default:None.
            activation(function): activation function. default:tf.nn.tanh.
        Returns:
            Returns the result of convolution of `inputs` and `laplacian_matrix`
        '''
        # [batch_size, num_node, num_feature]
        input_shape = inputs.get_shape().with_rank(3)

        num_node = tf.shape(inputs)[-2]
        num_feature = input_shape[-1].value

        # GC on inputs
        # reshape from [batch, num_node, num_feature] into [num_node, batch*num_feature]
        gc_input = tf.reshape(tf.transpose(inputs, perm=[1, 0, 2]), [num_node, -1])

        # Chebyshev polynomials
        # Reference: https://github.com/mdeff/cnn_graph
        gc_outputs = list()
        # Xt_0 = T_0 X = I X = X.
        gc_outputs.append(gc_input)
        # Xt_1 = T_1 X = L X.
        if gcn_k >= 1:
            gc_outputs.append(tf.matmul(laplacian_matrix, gc_input))
        # Xt_k = 2 L Xt_k-1 - Xt_k-2.
        for k in range(2, gcn_k+1):
            gc_outputs.append(2 * tf.matmul(laplacian_matrix, gc_outputs[-1]) - gc_outputs[-1])

        # [gcn_k+1, number_nodes, batch*num_feature]
        gc_outputs = tf.reshape(gc_outputs, [gcn_k+1, num_node, -1, num_feature])
        # [batch, number_nodes, num_feature, gcn_k+1]
        gc_outputs = tf.transpose(gc_outputs, [2, 1, 3, 0])
        # [batch*number_nodes, num_feature*gcn_k+1]
        gc_outputs = tf.reshape(gc_outputs, [-1, num_feature*(gcn_k+1)])

        output_weight = tf.get_variable("weights", shape=[num_feature*(gcn_k+1), output_size],
                                        trainable=trainable, dtype=dtype,
                                        initializer=initializer, regularizer=regularizer)
        gc_outputs = tf.matmul(gc_outputs, output_weight)

        if use_bias:
            biases = tf.get_variable("biases", [output_size], dtype=dtype,
                                     initializer=tf.constant_initializer(0, dtype=dtype))
            gc_outputs = tf.nn.bias_add(gc_outputs, biases)

        gc_outputs = tf.reshape(gc_outputs, [-1, num_node, output_size])

        return activation(gc_outputs)

    @staticmethod
    def add_multi_gc_layers(inputs, gcn_k, gcn_l, output_size, laplacian_matrix, activation=tf.nn.tanh):
        '''
        Call add_gc_layer function to add multi Graph Convolution Layer.`gcn_l` is the number of layers added.
        '''
        with tf.variable_scope('multi_gcl', reuse=False):
            for i in range(gcn_l):
                with tf.variable_scope('gcl_%s' % i, reuse=False):
                    inputs = GCL.add_gc_layer(inputs=inputs,
                                              gcn_k=gcn_k,
                                              laplacian_matrix=laplacian_matrix,
                                              output_size=output_size,
                                              activation=activation)
        return inputs



import os

from multiprocessing import Pool, Manager
from functools import reduce


# (my_rank, n_jobs, dataList, resultHandleFunction, parameterList)
def multiple_process(distribute_list, partition_func, task_func, n_jobs, reduce_func, parameters):

    """
    Args:

        distribute_list(list): The "data" list to be partitioned, such as a list of files which will be
            distributed among different tasks and each task process a part of the files.
        partition_func(function): Partition function will be used to cut the distribute_list, it should accept
            three inputs: distribute_list, i, n_job, where i is the index of jobs (i.e. integer from 0 to n_jobs-1),
            n_jos is the number of threads; partition function should return a data_list for the job_i
        task_func(function): Task function, where the inputs are share_queue, locker, data, parameters, no return.
            pls refer to the DiDi-Data processing codes for more information.
        n_jobs(int): Number of threads
        reduce_func(function): Reduce function which combine the outputs from all the threads into one final output.
        parameters(list): parameters send to the task function

    """

    if callable(partition_func) and callable(task_func) and callable(reduce_func):
        print('Parent process %s.' % os.getpid())

        manager = Manager()
        share_queue = manager.Queue()
        locker = manager.Lock()

        p = Pool()
        for i in range(n_jobs):
            p.apply_async(task_func, args=(share_queue, locker, partition_func(distribute_list, i, n_jobs),
                                           [i] + parameters,))
        print('Waiting for all sub_processes done...')
        p.close()
        p.join()
        print('All sub_processes done.')

        result_list = []
        while not share_queue.empty():
            result_list.append(share_queue.get_nowait())

        return reduce(reduce_func, result_list)
    else:
        print('Parameter error')


"""
# Example
def task(share_queue, locker, data, parameters):

    print('Child process %s with pid %s' % (parameters[0], os.getpid()))

    result = sum(data)

    locker.acquire()
    share_queue.put(result)
    locker.release()


if __name__ == "__main__":

    data = [e for e in range(1000000)]

    n_job = 4

    sum_result = \
        multiple_process(distribute_list=data,
                         partition_func=lambda data, i, n_job: [data[e] for e in range(len(data)) if e % n_job == i],
                         task_func=task, n_jobs=n_job, reduce_func=lambda x, y: x + y, parameters=[])

    print('Result', sum_result)
"""


import keras
import tensorflow as tf

from ..model_unit import BaseModel
from ..model_unit import GAL, GCL
from ..model_unit import DCGRUCell
from ..model_unit import GCLSTMCell


class STMeta(BaseModel):
    """
        Args:
            num_node(int): Number of nodes in the graph, e.g. number of stations in NYC-Bike dataset.
            external_dim(int): Dimension of the external feature, e.g. temperature and wind are two dimension.
            closeness_len(int): The length of closeness data history. The former consecutive ``closeness_len`` time slots
            of data will be used as closeness history.
            period_len(int): The length of period data history. The data of exact same time slots in former consecutive
            ``period_len`` days will be used as period history.
            trend_len(int): The length of trend data history. The data of exact same time slots in former consecutive
            ``trend_len`` weeks (every seven days) will be used as trend history.
            input_dim(int): The dimension of input features. 1 if "with_tpe" (data_loader parameters) = False, otherwise 0.
            num_graph(int): Number of graphs used in STMeta.
            gcn_k(int): The highest order of Chebyshev Polynomial approximation in GCN.
            gcn_layers(int): Number of GCN layers.
            gclstm_layers(int): Number of STRNN layers, it works on all modes of STMeta such as GCLSTM and DCRNN.
            num_hidden_units(int): Number of hidden units of RNN.
            num_dense_units(int): Number of dense units.
            graph_merge_gal_units(int): Number of units in GAL for merging different graph features.
                Only works when graph_merge='gal'
            graph_merge_gal_num_heads(int): Number of heads in GAL for merging different graph features.
                Only works when graph_merge='gal'
            temporal_merge_gal_units(int): Number of units in GAL for merging different temporal features.
                Only works when temporal_merge='gal'
            temporal_merge_gal_num_heads(int): Number of heads in GAL for merging different temporal features.
                Only works when temporal_merge='gal'
            st_method(str): must in ['GCLSTM', 'DCRNN', 'GRU', 'LSTM'], which refers to different
                spatial-temporal modeling methods.
                'GCLSTM': GCN for modeling spatial feature, LSTM for modeling temporal feature.
                'DCRNN': Diffusion Convolution for modeling spatial feature, GRU for modeling temporam frature.
                'GRU': Ignore the spatial, and model the temporal feature using GRU
                'LSTM': Ignore the spatial, and model the temporal feature using LSTM
            temporal_merge(str): must in ['gal', 'concat'], refers to different temporal merging methods,
                'gal': merge using GAL,
                'concat': merge by concat and dense
            graph_merge(str): must in ['gal', 'concat'], refers to different graph merging methods,
                'gal': merge using GAL,
                'concat': merge by concat and dense
            output_activation(function): activation function, e.g. tf.nn.tanh
            lr(float): Learning rate. Default: 1e-5
            code_version(str): Current version of this model code, which will be used as filename for saving the model
            model_dir(str): The directory to store model files. Default:'model_dir'.
            gpu_device(str): To specify the GPU to use. Default: '0'.
        """

    def __init__(self,
                 num_node,
                 external_dim,
                 closeness_len,
                 period_len,
                 trend_len,
                 input_dim=1,

                 # gcn parameters
                 num_graph=1,
                 gcn_k=1,
                 gcn_layers=1,
                 gclstm_layers=1,

                 # dense units
                 num_hidden_units=64,
                 # LSTM units
                 num_dense_units=32,

                 # merge parameters
                 graph_merge_gal_units=32,
                 graph_merge_gal_num_heads=2,
                 temporal_merge_gal_units=64,
                 temporal_merge_gal_num_heads=2,

                 # network structure parameters
                 st_method='GCLSTM',  # gclstm
                 temporal_merge='gal',  # gal
                 graph_merge='gal',  # concat

                 output_activation=tf.nn.sigmoid,

                 lr=1e-4,
                 code_version='STMeta-QuickStart',
                 model_dir='model_dir',
                 gpu_device='0', **kwargs):

        super(STMeta, self).__init__(code_version=code_version, model_dir=model_dir, gpu_device=gpu_device)

        self._num_node = num_node
        self._input_dim = input_dim
        print("self._input_dim",self._input_dim)
        self._gcn_k = gcn_k
        self._gcn_layer = gcn_layers
        self._graph_merge_gal_units = graph_merge_gal_units
        self._graph_merge_gal_num_heads = graph_merge_gal_num_heads
        self._temporal_merge_gal_units = temporal_merge_gal_units
        self._temporal_merge_gal_num_heads = temporal_merge_gal_num_heads
        self._gclstm_layers = gclstm_layers
        self._num_graph = num_graph
        self._external_dim = external_dim
        self._output_activation = output_activation

        self._st_method = st_method.upper()
        self._temporal_merge = temporal_merge
        self._graph_merge = graph_merge

        self._closeness_len = int(closeness_len)
        self._period_len = int(period_len)
        self._trend_len = int(trend_len)
        self._num_hidden_unit = num_hidden_units
        self._num_dense_units = num_dense_units
        self._lr = lr
    
    def build(self, init_vars=True, max_to_keep=5):
        with self._graph.as_default():

            temporal_features = []

            if self._closeness_len is not None and self._closeness_len > 0:
                closeness_feature = tf.placeholder(tf.float32, [None, None, self._closeness_len, self._input_dim],
                                                   name='closeness_feature')
                self._input['closeness_feature'] = closeness_feature.name
                temporal_features.append([self._closeness_len, closeness_feature, 'closeness_feature'])

            if self._period_len is not None and self._period_len > 0:
                period_feature = tf.placeholder(tf.float32, [None, None, self._period_len, self._input_dim],
                                                name='period_feature')
                self._input['period_feature'] = period_feature.name
                temporal_features.append([self._period_len, period_feature, 'period_feature'])

            if self._trend_len is not None and self._trend_len > 0:
                trend_feature = tf.placeholder(tf.float32, [None, None, self._trend_len, self._input_dim],
                                               name='trend_feature')
                self._input['trend_feature'] = trend_feature.name
                temporal_features.append([self._trend_len, trend_feature, 'trend_feature'])

            if len(temporal_features) > 0:
                target = tf.placeholder(tf.float32, [None, None, 1], name='target')
                laplace_matrix = tf.placeholder(tf.float32, [self._num_graph, None, None], name='laplace_matrix')
                self._input['target'] = target.name
                self._input['laplace_matrix'] = laplace_matrix.name
            else:
                raise ValueError('closeness_len, period_len, trend_len cannot all be zero')

            graph_outputs_list = []

            for graph_index in range(self._num_graph):

                if self._st_method in ['GCLSTM', 'DCRNN', 'GRU', 'LSTM']:

                    outputs_temporal = []

                    for time_step, target_tensor, given_name in temporal_features:

                        if self._st_method == 'GCLSTM':

                            multi_layer_cell = tf.keras.layers.StackedRNNCells(
                                [GCLSTMCell(units=self._num_hidden_unit, num_node=self._num_node,
                                            laplacian_matrix=laplace_matrix[graph_index],
                                            gcn_k=self._gcn_k, gcn_l=self._gcn_layer)
                                 for _ in range(self._gclstm_layers)])

                            outputs = tf.keras.layers.RNN(multi_layer_cell)(tf.reshape(target_tensor, [-1, time_step, self._input_dim]))

                            st_outputs = tf.reshape(outputs, [-1, 1, self._num_hidden_unit])

                        elif self._st_method == 'DCRNN':
                        # laplace_matrix will be diffusion_matrix when self._st_method == 'DCRNN'

                            encoding_cells = [DCGRUCell(self._num_hidden_unit, self._input_dim, self._num_graph,
                                             laplace_matrix,
                                             max_diffusion_step=self._gcn_k,
                                             num_node=self._num_node, name=str(graph_index) + given_name) for _ in range(self._gclstm_layers)]

                            encoding_cells = tf.contrib.rnn.MultiRNNCell(encoding_cells, state_is_tuple=True)

                            inputs_unstack = tf.unstack(tf.reshape(target_tensor, [-1, self._num_node, time_step]),
                                                        axis=-1)

                            outputs, _ = \
                                tf.contrib.rnn.static_rnn(encoding_cells, inputs_unstack, dtype=tf.float32)

                            st_outputs = tf.reshape(outputs[-1], [-1, 1, self._num_hidden_unit])

                        elif self._st_method == 'GRU':

                            multi_layer_gru = tf.keras.layers.StackedRNNCells([tf.keras.layers.GRUCell(units=self._num_hidden_unit) for _ in range(self._gclstm_layers)])
                            
                            outputs = tf.keras.layers.RNN(multi_layer_gru)(
                                tf.reshape(target_tensor, [-1, time_step, self._input_dim]))
                            st_outputs = tf.reshape(outputs, [-1, 1, self._num_hidden_unit])

                        elif self._st_method == 'LSTM':

                            multi_layer_gru = tf.keras.layers.StackedRNNCells([tf.keras.layers.LSTMCell(units=self._num_hidden_unit) for _ in range(self._gclstm_layers)])
                            
                            outputs = tf.keras.layers.RNN(multi_layer_gru)(
                                tf.reshape(target_tensor, [-1, time_step, self._input_dim]))
                            st_outputs = tf.reshape(outputs, [-1, 1, self._num_hidden_unit])

                        outputs_temporal.append(st_outputs)

                    if self._temporal_merge == 'concat':
                        
                        graph_outputs_list.append(tf.concat(outputs_temporal, axis=-1))

                    elif self._temporal_merge == 'gal':

                        _, gal_output = GAL.add_ga_layer_matrix(inputs=tf.concat(outputs_temporal, axis=-2),
                                                                units=self._temporal_merge_gal_units,
                                                                num_head=self._temporal_merge_gal_num_heads)

                        graph_outputs_list.append(tf.reduce_mean(gal_output, axis=-2, keepdims=True))

            if self._num_graph > 1:

                if self._graph_merge == 'gal':
                    # (graph, inputs_name, units, num_head, activation=tf.nn.leaky_relu)
                    _, gal_output = GAL.add_ga_layer_matrix(inputs=tf.concat(graph_outputs_list, axis=-2),
                                                            units=self._graph_merge_gal_units,
                                                            num_head=self._graph_merge_gal_num_heads)
                    dense_inputs = tf.reduce_mean(gal_output, axis=-2, keepdims=True)

                elif self._graph_merge == 'concat':

                    dense_inputs = tf.concat(graph_outputs_list, axis=-1)

            else:

                dense_inputs = graph_outputs_list[-1]

            dense_inputs = tf.reshape(dense_inputs, [-1, self._num_node, 1, dense_inputs.get_shape()[-1].value])

            dense_inputs = tf.keras.layers.BatchNormalization(axis=-1, name='feature_map')(dense_inputs)

            # external dims
            if self._external_dim is not None and self._external_dim > 0:
                external_input = tf.placeholder(tf.float32, [None, self._external_dim])
                self._input['external_feature'] = external_input.name
                external_dense = tf.keras.layers.Dense(units=10)(external_input)
                external_dense = tf.tile(tf.reshape(external_dense, [-1, 1, 1, 10]),
                                         [1, tf.shape(dense_inputs)[1], tf.shape(dense_inputs)[2], 1])
                dense_inputs = tf.concat([dense_inputs, external_dense], axis=-1)

            dense_output0 = tf.keras.layers.Dense(units=self._num_dense_units,
                                                  activation=tf.nn.tanh,
                                                  use_bias=True,
                                                  kernel_initializer='glorot_uniform',
                                                  bias_initializer='zeros',
                                                  kernel_regularizer=tf.keras.regularizers.l2(0.01),
                                                  bias_regularizer=tf.keras.regularizers.l2(0.01)
                                                  )(dense_inputs)

            dense_output1 = tf.keras.layers.Dense(units=self._num_dense_units,
                                                  activation=tf.nn.tanh,
                                                  use_bias=True,
                                                  kernel_initializer='glorot_uniform',
                                                  bias_initializer='zeros',
                                                  kernel_regularizer=tf.keras.regularizers.l2(0.01),
                                                  bias_regularizer=tf.keras.regularizers.l2(0.01)
                                                  )(dense_output0)

            pre_output = tf.keras.layers.Dense(units=1,
                                               activation=tf.nn.tanh,
                                               use_bias=True,
                                               kernel_initializer='glorot_uniform',
                                               bias_initializer='zeros',
                                               kernel_regularizer=tf.keras.regularizers.l2(0.01),
                                               bias_regularizer=tf.keras.regularizers.l2(0.01)
                                               )(dense_output1)

            prediction = tf.reshape(pre_output, [-1, self._num_node, 1], name='prediction')

            loss_pre = tf.sqrt(tf.reduce_mean(tf.square(target - prediction)), name='loss')

            train_op = tf.train.AdamOptimizer(self._lr).minimize(loss_pre, name='train_op')

            # record output
            self._output['prediction'] = prediction.name
            self._output['loss'] = loss_pre.name

            # record train operation
            self._op['train_op'] = train_op.name

        super(STMeta, self).build(init_vars, max_to_keep)

    # Define your '_get_feed_dict function‘, map your input to the tf-model
    def _get_feed_dict(self,
                       laplace_matrix,
                       closeness_feature=None,
                       period_feature=None,
                       trend_feature=None,
                       target=None,
                       external_feature=None):
        feed_dict = {
            'laplace_matrix': laplace_matrix,
        }
        if target is not None:
            feed_dict['target'] = target
        if self._external_dim is not None and self._external_dim > 0:
            feed_dict['external_feature'] = external_feature
        if self._closeness_len is not None and self._closeness_len > 0:
            feed_dict['closeness_feature'] = closeness_feature
        if self._period_len is not None and self._period_len > 0:
            feed_dict['period_feature'] = period_feature
        if self._trend_len is not None and self._trend_len > 0:
            feed_dict['trend_feature'] = trend_feature
        return feed_dict
